"""Kat model compiler.

This takes parsed productions, figures out the dependencies and builds the model in the
correct order.

Sean Leavey <sean.leavey@ligo.org>
"""

import logging
from collections import defaultdict, ChainMap
from io import StringIO
import re
from functools import singledispatchmethod
from networkx import (
    selfloop_edges,
    lexicographical_topological_sort,
    simple_cycles,
    descendants,
    NetworkXUnfeasible,
)
import numpy as np
from .. import Model
from ..symbols import Resolving
from ..utilities import option_list, ngettext
from ..exceptions import (
    FinesseException,
    ModelAttributeError,
    ModelParameterDefaultValueError,
    ModelParameterSelfReferenceError,
)
from . import syntax
from .graph import KatGraph, KatNodeType, KatEdgeType
from .containers import (
    KatToken,
    KatReferenceToken,
    KatScript,
    KatElement,
    KatFunction,
    KatKwarg,
    KatGroupedExpression,
    KatExpression,
    KatArray,
    KatNumericalArray,
)
from .parser import KatParser
from .exceptions import KatScriptError, KatSyntaxError, KatMissingAfterDirective
from .util import duplicates, merge_attributes

LOGGER = logging.getLogger(__name__)


def production(node, graph):
    from .generator import KatUnfiller

    unfiller = KatUnfiller()
    return unfiller.production(node, graph)


def argument_signature_parameter(argument_node, adapter, graph):
    data = graph.nodes[argument_node]
    parameters = adapter.call_signature(exclude_name=True).parameters

    if key_token := data.get("key_token"):
        # It's a keyword argument: just look up the key.
        return parameters[key_token.value]

    # It's a positional argument: look up the correct index.
    return list(parameters.values())[graph.argument_node_order(argument_node)]


class KatCompiler:
    """Kat model compiler."""

    def __init__(self, spec=None):
        if spec is None:
            from .spec import KatSpec

            spec = KatSpec()

        self.spec = spec
        self.graph = None
        self.build_graph = None
        self._parser = None
        self._parameter_dependencies = None

    @property
    def script(self):
        return self._parser.script

    def compile(self, string, **kwargs):
        """Compile the contents of `string`.

        Parameters
        ----------
        string : :class:`str`
            The string to compile kat script from.

        Other Parameters
        ----------------
        kwargs
            Keyword arguments supported by :meth:`.compile_file`.

        Returns
        -------
        :class:`.Model`
            The model compiled from reading `string`.
        """
        return self.compile_file(StringIO(string), **kwargs)

    def compile_file(self, fobj, model=None, resolve=True, build=True):
        """Compile the contents of the specified file.

        Parameters
        ----------
        fobj : :class:`io.FileIO`
            The file object to compile kat script from. This should be opened in text
            mode.

        model : :class:`.Model`, optional
            An existing model to compile the contents of `fobj` into. If not specified,
            a new, empty model is created.

        resolve : :class:`bool`, optional
            Resolve the parsed script. If False, the parsed contents is added to graph
            but no sanity checks are performed and nothing is returned. Defaults to
            True.

        build : :class:`bool`, optional
            Build the parsed script. If False, no Finesse objects will be added to the
            model and nothing will be returned. Defaults to True.

        Returns
        -------
        :class:`.Model` or None
            If `resolve` and `build` are True, the model compiled from reading `fobj`;
            None otherwise.

        Raises
        ------
        :class:`.KatSyntaxError`
            If a syntax error is present in the contents of `fobj`.
        """
        # Reset state.
        self._parser = KatParser()
        self.graph = KatGraph()
        self.build_graph = None
        self._parameter_dependencies = []

        try:
            script = self._parser.parse_file(fobj)
        except KatSyntaxError as e:
            # There was a syntax error. Try to improve the error message using the spec.
            self._reraise_parser_error_in_spec_context(e)

        self._fill(script, self.graph.ROOT_NODE_NAME)

        # At this stage there shouldn't be any dependencies between branches.
        assert self.graph.is_tree()

        if resolve:
            self._resolve()

            if build:
                return self._build(self.graph.ROOT_NODE_NAME, model)

    @property
    def _directive_functions(self):
        return ChainMap(self.spec.commands, self.spec.analyses)

    @property
    def _expression_functions(self):
        return ChainMap(self.spec.expression_functions, self.spec.unary_operators)

    @property
    def _available_functions(self):
        return ChainMap(self._directive_functions, self._expression_functions)

    def _fill(self, value, path, **attributes):
        assert path not in self.graph
        attributes = merge_attributes(
            attributes, self._item_node_attributes(value, path)
        )
        self.graph.add_node(path, **attributes)

        if hasattr(value, "arguments"):
            # Assemble arguments.
            for order, argument in enumerate(value.arguments):
                argument_path = self.graph.item_node_name(order, path)
                self._fill(argument, argument_path)
                self.graph.add_edge(
                    argument_path, path, type=KatEdgeType.ARGUMENT, order=order
                )

        return attributes

    @singledispatchmethod
    def _item_node_attributes(self, item, path):
        """Get the attributes for `item` to be added to the corresponding node."""
        raise NotImplementedError(
            f"don't know how to handle an item with type '{item.__class__.__name__}'"
        )

    @_item_node_attributes.register(KatScript)
    def _(self, script, path):
        return {"type": KatNodeType.ROOT, "extra_tokens": script.extra}

    @_item_node_attributes.register(KatToken)
    def _(self, token, path):
        attributes = {"token": token}
        if token.type == "NAME":
            # This could be a keyword, constant or parameter reference.
            name = token.value

            if name in self.spec.keywords:
                token_type = KatNodeType.KEYWORD
            elif name in self.spec.constants:
                token_type = KatNodeType.CONSTANT
            else:
                # Assume this is a parameter like `l1.P`.
                self._parameter_dependencies.append(path)
                token_type = KatNodeType.PARAMETER
        else:
            # This is some other token like a number or string.
            token_type = KatNodeType.VALUE

        attributes["type"] = token_type

        return attributes

    @_item_node_attributes.register(KatReferenceToken)
    def _(self, token, path):
        attributes = {"token": token, "type": KatNodeType.PARAMETER_REFERENCE}
        self._parameter_dependencies.append(path)
        return attributes

    @_item_node_attributes.register(KatKwarg)
    def _(self, kwarg, path):
        return self._fill(
            kwarg.value, path, key_token=kwarg.key, extra_tokens=kwarg.extra
        )

    @_item_node_attributes.register(KatElement)
    def _(self, element, path):
        directive_token = element.directive
        element_type = directive_token.value
        if element_type not in self.spec.elements:
            if element_type in self.spec.commands or element_type in self.spec.analyses:
                # Syntax error was on an command or analysis definition.
                raise KatWrongSyntaxError(directive_token, self.script, self.spec)
            else:
                raise KatUnrecognisedElementError(element, self.script)
        return {
            "token": element.directive,
            "name_token": element.name,
            "type": KatNodeType.ELEMENT,
            "extra_tokens": element.extra,
        }

    @_item_node_attributes.register(KatFunction)
    def _(self, function, path):
        directive_token = function.directive
        if directive_token.value not in self._available_functions:
            raise KatUnrecognisedFunctionError(function, self.script)
        return {
            "token": directive_token,
            "type": KatNodeType.FUNCTION,
            "extra_tokens": function.extra,
        }

    @_item_node_attributes.register(KatGroupedExpression)
    def _(self, group, path):
        return {
            "type": KatNodeType.GROUPED_EXPRESSION,
            "extra_tokens": group.extra,
        }

    @_item_node_attributes.register(KatExpression)
    def _(self, expression, path):
        operator_token = expression.operator
        if operator_token.value not in self.spec.binary_operators:
            raise KatUnrecognisedTypeError(
                f"unknown expression operator '{operator_token.raw_value}'",
                self.script,
                expression.operator,
            )
        return {"type": KatNodeType.EXPRESSION, "token": operator_token}

    @_item_node_attributes.register(KatArray)
    def _(self, array, path):
        return {"type": KatNodeType.ARRAY, "extra_tokens": array.extra}

    @_item_node_attributes.register(KatNumericalArray)
    def _(self, array, path):
        return {"type": KatNodeType.NUMERICAL_ARRAY, "extra_tokens": array.extra}

    def _resolve(self):
        """Resolve references and perform sanity checks on the graph.

        Dependencies due to parameters (e.g. `m1.p1.o`) and parameter references (e.g.
        `&l1.P`) are created by drawing edges to the relevant nodes, transforming the
        parse tree into a graph.

        This additionally checks that:

        - element names are valid and unique,
        - keyword arguments aren't defined in duplicate, and
        - no cycles exist due to dependencies (i.e. if a reference depends on the
          referencing component or its ancestors in any way).

        The `order` attributes of the root edges are overwritten using a topological
        sort to ensure the resulting graph is built in a way that most types of
        dependency can be resolved.
        """
        # Various types of node that we need to validate.
        elements = defaultdict(list)
        singular_directives = defaultdict(list)
        analyses = []

        for node, data in self.graph.root_directive_nodes(data=True):
            node_type = data["type"]
            token = data["token"]
            directive = token.value

            if node_type in KatNodeType.DIRECTIVE_NODES:
                if node_type in KatNodeType.ELEMENT:
                    name_token = data["name_token"]
                    name = name_token.value

                    if name in self.spec.constants:
                        # Disallow.
                        raise KatScriptError(
                            f"constant '{name}' cannot be used as an element name",
                            self.script,
                            [[name_token]],
                        )
                    elif name in self.spec.keywords:
                        LOGGER.warning(
                            f"element name '{name}' (line {name_token.lineno}) is also "
                            f"the name of a keyword, which may lead to confusion"
                        )

                    adapter = self._element_adapter(directive)
                    elements[name].append(node)
                elif node_type in KatNodeType.FUNCTION:
                    adapter = self._function_adapter(directive)
                    if directive in self.spec.analyses:
                        analyses.append(token)

                if adapter.singular:
                    singular_directives[directive].append(node)

                # Check for duplicate keyword arguments.
                dupekvs = duplicates(
                    [
                        self.graph.nodes[node]["key_token"]
                        for node in self.graph.dependent_argument_nodes(node)
                        if "key_token" in self.graph.nodes[node]
                    ],
                    key=lambda token: token.raw_value,
                )
                if dupekvs:
                    keys, dupetokens = zip(*dupekvs)
                    keylist = option_list([f"'{key}'" for key in keys], final_sep="and")
                    msg = ngettext(
                        len(keys),
                        f"duplicate arguments with key {keylist}",
                        f"duplicate arguments with keys {keylist}",
                        sub=False,
                    )
                    error_tokens = [
                        [token] for tokens in dupetokens for token in tokens
                    ]
                    raise KatScriptError(msg, self.script, error_tokens)

        for directive, nodes in singular_directives.items():
            if len(nodes) > 1:
                raise KatScriptError(
                    f"there can only be one '{directive}' directive",
                    self.script,
                    [[self.graph.nodes[node]["token"]] for node in nodes],
                )

        for name, nodes in elements.items():
            if len(nodes) > 1:
                raise KatScriptError(
                    f"multiple elements with name '{name}'",
                    self.script,
                    [[self.graph.nodes[node]["name_token"]] for node in nodes],
                )

        # Check that there isn't more than one root analysis.
        if len(analyses) > 1:
            raise KatScriptError(
                "duplicate analysis trees (combine with 'series' or 'parallel')",
                self.script,
                [[token] for token in analyses],
            )

        # Create dependencies for parameters and parameter references.
        for source_param_path in self._parameter_dependencies:
            source_token = self.graph.nodes[source_param_path]["token"]

            # Targets may not exist in the current script; they may already be in the
            # model prior to parsing.
            try:
                target_directive_path = self.graph.param_target_element_path(
                    source_token.value
                )
            except ValueError:
                LOGGER.debug(
                    f"parameter '{source_token.raw_value}' targetted by "
                    f"'{source_param_path}' does not exist in the script - ignoring"
                )
                continue

            # Add the dependency.
            self.graph.add_edge(
                target_directive_path, source_param_path, type=KatEdgeType.DEPENDENCY,
            )

        # Create graph with just the root directives (use a copy because it may be
        # modified during compilation).
        self.build_graph = self.graph.root_directive_graph()

        def node_build_order(source_graph):
            """Compute node topological order, or throw a cycle error."""
            # Remove self-references, since these will be dealt with by the compiler.
            graph = source_graph.copy()
            for edge in selfloop_edges(source_graph):
                graph.remove_edge(*edge)

            try:
                # Compute topological order, resolving ambiguities in favour of lower
                # line numbers.
                # Convert to list so we know immediately if there are cycles.
                return list(
                    lexicographical_topological_sort(
                        graph, key=lambda node: graph.nodes[node]["token"].lineno
                    )
                )
            except NetworkXUnfeasible:
                # The graph contains at least one cyclic dependency. Work out the
                # cause(s) by finding cycles using the full graph.
                cyclic_param_nodes = set()
                for cycle_nodes in simple_cycles(self.graph):
                    for cycle_node in cycle_nodes:
                        data = self.graph.nodes[cycle_node]
                        if data["type"] not in KatNodeType.DEPENDENT_NODES:
                            # Ignore anything in the cycle that isn't a reference node
                            # (e.g. directives).
                            continue

                        cyclic_param_nodes.add(cycle_node)

                # Grab and sort the cyclic tokens in line order.
                cyclic_parameter_tokens = sorted(
                    [self.graph.nodes[node]["token"] for node in cyclic_param_nodes],
                    key=lambda tok: tok.lineno,
                )

                raise KatCycleError(self.script, cyclic_parameter_tokens)

        # Split the nodes into groups depending on whether they descend from a node with
        # the "build_last" flag.
        first_build_nodes = set(self.build_graph.nodes)
        second_build_nodes = set()
        for node in self.build_graph.nodes():
            adapter = self._directive_adapter(self.graph.nodes[node]["token"].value)
            if not adapter.build_last or node in second_build_nodes:
                # Keep the directive in the first set.
                continue

            # All nodes that descend from (depend on) specified node.
            children = descendants(self.build_graph, node)

            # Move node and descendants from first to second build step.
            first_build_nodes.remove(node)
            first_build_nodes.difference_update(children)
            second_build_nodes.add(node)
            second_build_nodes.update(children)

        first_build_subgraph = self.build_graph.subgraph(first_build_nodes)
        second_build_subgraph = self.build_graph.subgraph(second_build_nodes)
        final_build_order = node_build_order(first_build_subgraph) + node_build_order(
            second_build_subgraph
        )

        # Reorder the root edges topologically.
        for source, target in self.graph.root_argument_edges():
            index = final_build_order.index(source)
            LOGGER.debug(f"{source} has build order {index}")
            self.graph.edges[(source, target)]["order"] = index

    def _build(self, node, model):
        data = self.graph.nodes[node]
        nodetype = data["type"]

        if nodetype in KatNodeType.COMPILER_TERMINAL_NODES:
            # Just use the value.
            value = data["token"].value
        elif nodetype == KatNodeType.CONSTANT:
            # Look up the symbol corresponding to the constant.
            value = self.spec.constants[data["token"].value]
        elif nodetype == KatNodeType.PARAMETER:
            # Copy existing model parameter by value.
            target = data["token"].value
            value = model.reduce_get_attr(target)
            LOGGER.debug(f"parameter {node} targets {target} by value")
        elif nodetype == KatNodeType.PARAMETER_REFERENCE:
            source_directive_path = self.graph.branch_base(node)

            if self.graph.has_edge(source_directive_path, node):
                # This is a reference to another parameter in the same element. Throw an
                # exception that can get caught by the argument compiler. Doing it this
                # way ensures expressions with nested self-references (like `1-&m1.T` in
                # `m m1 1-&m1.T 0`) get set as resolving.
                LOGGER.debug(f"{node} is a self-reference")
                raise KatParameterSelfReferenceException()

            # Copy existing model parameter by reference.
            target = data["token"].value
            try:
                value = model.reduce_get_attr(target).ref
            except FinesseException as e:
                # There was an error getting a reference to the target.
                raise KatParameterCompilationError(
                    e, self.script, node, self.graph
                ) from e
            LOGGER.debug(f"parameter {node} targets {target} by reference")
        elif nodetype == KatNodeType.EXPRESSION:
            operator = self.spec.binary_operators[data["token"].value]
            arguments = self._built_arguments(node, model)
            # Arguments are already in order.
            assert len(arguments) == 2
            lhs, rhs = arguments
            value = operator(lhs, rhs)

            # Eagerly evaluate the expression if its arguments don't depend on anything
            # else.
            if self.graph.is_independent(node):
                value = value.eval()
                LOGGER.debug(
                    f"eagerly evaluated dependencyless expression '{node}' to a "
                    f"{type(value)}"
                )
        elif nodetype == KatNodeType.GROUPED_EXPRESSION:
            arguments = self._built_arguments(node, model)
            assert len(arguments) == 1
            value = arguments[0]
        elif nodetype == KatNodeType.ARRAY:
            value = self._built_arguments(node, model)
        elif nodetype == KatNodeType.NUMERICAL_ARRAY:
            value = np.array(self._built_arguments(node, model))
        elif nodetype == KatNodeType.ELEMENT:
            args, kwargs = self._built_directive_params(node, model)
            adapter = self._element_adapter(data["token"].value)

            # Add name to arguments.
            name = data["name_token"].value
            args = [name, *args]

            try:
                value = adapter.compile(args, kwargs)
            except Exception as e:
                raise KatDirectiveCompilationError(
                    e, self.script, node, adapter, self.graph, self.spec
                ) from e
        elif nodetype == KatNodeType.FUNCTION:
            function_name = data["token"].value

            if function_name in self._expression_functions:
                operator = self._expression_functions[function_name]
                args, kwargs = self._built_directive_params(node, model)

                try:
                    value = operator(*args, **kwargs)
                except TypeError as e:
                    # Replace 'lambda' in the function init error with the function
                    # name.
                    args = list(e.args)
                    args[0] = re.sub(
                        r"(\<lambda\>(\d+)?|\w+)\(\)", f"'{function_name}'", args[0],
                    )
                    raise TypeError(*args)

                # Eagerly evaluate the function if its arguments don't depend on
                # anything else.
                if self.graph.is_independent(node):
                    value = value.eval()
                    LOGGER.debug(
                        f"eagerly evaluated dependencyless function '{node}' to a "
                        f"{type(value)}"
                    )
            else:
                # A directive function.
                adapter = self._function_adapter(data["token"].value)
                args, kwargs = self._built_directive_params(node, model)

                if hasattr(adapter, "compile"):
                    try:
                        value = adapter.compile(args, kwargs)
                    except Exception as e:
                        raise KatDirectiveCompilationError(
                            e, self.script, node, adapter, self.graph, self.spec
                        ) from e
                else:
                    LOGGER.debug(f"function {node} has arguments {args}, {kwargs}")
                    value = args, kwargs
        elif nodetype == KatNodeType.ROOT:
            if model is None:
                LOGGER.debug("creating new model")
                model = Model()
            else:
                LOGGER.debug(f"building into existing model {model!r}")

            for argument_node in self._sorted_arguments(node):
                item = self._build(argument_node, model)
                directive = self.graph.nodes[argument_node]["token"].value
                adapter = self._directive_adapter(directive)

                LOGGER.debug(f"applying {item!r} to {model!r}")
                try:
                    adapter.apply(model, item)
                except Exception as e:
                    raise KatDirectiveCompilationError(
                        e, self.script, argument_node, adapter, self.graph, self.spec
                    ) from e

                # Get all nodes connected to this argument node by a DEPENDENCY edge,
                # where the node is a subbranch of the argument node.
                self_references = self.graph.filter_dependent_nodes(
                    argument_node,
                    key=lambda refnode, _: self.graph.branch_base(refnode)
                    == argument_node,
                )

                if self_references:
                    # Convert to a list because we'll iterate multiple times.
                    self_references = list(self_references)

                    # Delete graph dependencies so that references resolve when we build
                    # again.
                    for target in self_references:
                        LOGGER.debug(f"breaking dependency {argument_node} -> {target}")
                        self.graph.remove_edge(argument_node, target)

                    for self_ref_path in self_references:
                        self_ref_param_path = self.graph.branch_base(
                            self_ref_path, argument_node
                        )

                        # Figure out the signature parameter's name and use it to grab
                        # the corresponding element's parameter. We have to do it this
                        # way because element parameters are unordered, and we might be
                        # dealing with positional arguments.
                        sigparam = argument_signature_parameter(
                            self_ref_param_path, adapter, self.graph
                        )
                        param = next(
                            param
                            for param in item.parameters
                            if param.name == sigparam.name
                        )

                        LOGGER.debug(f"resolving self-referencing {param.full_name}")
                        try:
                            value = self._build(self_ref_param_path, model)
                        except Exception as e:
                            raise KatParameterCompilationError(
                                e, self.script, self_ref_param_path, self.graph
                            ) from e

                        # Get rid of keys.
                        if isinstance(value, dict):
                            assert len(value) == 1
                            _, value = value.popitem()

                        LOGGER.debug(f"setting {param.full_name} to {type(value)}")

                        try:
                            param.value = value
                        except Exception as e:
                            raise KatParameterCompilationError(
                                e, self.script, self_ref_param_path, self.graph
                            ) from e

                    LOGGER.debug(f"finished resolving self-refs for {item!r}")

            value = model
        else:
            raise RuntimeError(f"don't know how to compile parameter '{nodetype}'")

        # Detect kwargs.
        if "key_token" in data:
            value = {data["key_token"].value: value}

        LOGGER.debug(f"compiled {nodetype} {node} to {type(value)}")
        return value

    def _sorted_arguments(self, node):
        return sorted(
            self.graph.dependent_argument_nodes(node),
            key=self.graph.argument_node_order,
        )

    def _built_arguments(self, node, model):
        """Get built dependent arguments of the current node."""
        return [
            self._build(argument_node, model)
            for argument_node in self._sorted_arguments(node)
        ]

    def _built_directive_params(self, node, model):
        """Get built dependent arguments of the current node in Python signature form.

        Self-referencing parameters are caught here too, and set to :class:`.Resolving`;
        these are fully resolved at the end of compilation.
        """
        args = []
        kwargs = {}
        for argument_node in self._sorted_arguments(node):
            try:
                item = self._build(argument_node, model)
            except KatParameterSelfReferenceException:
                item = Resolving()

                # Detect kwargs.
                if key_token := self.graph.nodes[argument_node].get("key_token"):
                    item = {key_token.value: item}

            if isinstance(item, dict):
                kwargs.update(item)
            else:
                args.append(item)
        return args, kwargs

    def _element_adapter(self, element):
        if element := self.spec.elements.get(element):
            return element

        raise ValueError(f"could not find element corresponding to '{element}'")

    def _function_adapter(self, function):
        if command := self.spec.commands.get(function):
            return command
        if analysis := self.spec.analyses.get(function):
            return analysis

        raise ValueError(f"could not find function corresponding to '{function}'")

    def _directive_adapter(self, directive):
        try:
            return self._element_adapter(directive)
        except ValueError:
            try:
                return self._function_adapter(directive)
            except ValueError:
                raise ValueError(
                    f"could not find directive corresponding to '{directive}'"
                )

    def _reraise_parser_error_in_spec_context(self, error):
        """Try to improve parser errors using the :class:`.KatSpec`.

        The parser is unaware of the available elements and functions etc. in the kat
        language spec, because this is only used during compilation. In some cases the
        error messages can be improved with knowledge of the spec, so this method
        identifies such cases.
        """
        if isinstance(error, KatMissingAfterDirective):
            # Use the spec to figure out whether the user might have missed a
            # parenthesis or a name.
            if error.directive.value in self.spec.directives:
                raise KatWrongSyntaxError(error.directive, self.script, self.spec)
            else:
                # Unrecognised directive type. Since that's the earlier error, throw
                # that.
                raise KatScriptError(
                    f"unknown element or function '{error.directive.raw_value}'",
                    self.script,
                    [[error.directive]],
                )

        # Nothing matched, so just reraise.
        raise error


class KatUnrecognisedTypeError(KatScriptError):
    """Error representing an unrecognised type."""

    def __init__(self, message, script, item):
        super().__init__(message, script, [[item]])


class KatUnrecognisedElementError(KatUnrecognisedTypeError):
    """Error representing an unrecognised element."""

    def __init__(self, element, script):
        directive = element.directive
        self.element_type = directive.raw_value
        super().__init__(f"unknown element '{self.element_type}'", script, directive)


class KatUnrecognisedFunctionError(KatUnrecognisedTypeError):
    """Error representing an unrecognised function."""

    def __init__(self, function, script):
        directive = function.directive
        self.function_type = directive.raw_value
        super().__init__(f"unknown function '{self.function_type}'", script, directive)


class KatWrongSyntaxError(KatSyntaxError):
    """Error representing an element, command or analysis that uses the wrong syntax."""

    def __init__(self, directive, script, spec):
        form = syntax(directive.raw_value, spec=spec, optional_as_positional=True)

        super().__init__(
            f"'{directive.raw_value}' should be written in the form '{form}'",
            script,
            directive,
        )


class KatBuildError(KatScriptError):
    """Error during the parser build step."""


class KatCycleError(KatBuildError):
    """Cyclic parameters."""

    def __init__(self, script, cyclic_nodes):
        # Convert list of error tokens to error item lists.
        items = [[item] for item in cyclic_nodes]
        super().__init__("cyclic parameters", script, items)


class KatDirectiveCompilationError(KatBuildError):
    def __init__(self, msg, script, statement_node, adapter, graph, spec):
        data = graph.nodes[statement_node]
        show_item = data["token"]
        error_items = [show_item]
        directive = show_item.value

        if msg:
            # Convert exceptions.
            msg = str(msg)

            # Replace the called Finesse Python API function name, if present, with the
            # directive name.
            msg = re.sub(r"(\<lambda\>|\w+)\(\)", f"'{directive}'", msg,)

            # Handle specific error messages.
            if matches := re.search(r"unexpected keyword argument '(\w+)'", msg):

                def keyfunc(_, data):
                    try:
                        return data["key_token"].value == matches.group(1)
                    except KeyError:
                        return False

                kwarg_node = next(graph.filter_argument_nodes(statement_node, keyfunc))
                # Mark the matching kwarg.
                item = production(kwarg_node, graph)
                # Mark the key.
                error_items = [item.key]

                # Add suggested syntax.
                expected = syntax(directive, spec=spec)
                msg = f"{msg} (expected syntax: {expected})"

        super().__init__(msg, script, [error_items])


class KatParameterCompilationError(KatBuildError):
    def __init__(self, error, script, param_node, graph):
        data = graph.nodes[param_node]
        value_token = data["token"]
        show_items = [value_token]

        # Add the key if it exists.
        try:
            show_items.append(data["key_token"])
        except KeyError:
            pass

        msg = str(error).strip()

        if isinstance(error, ModelAttributeError):
            # Mark only the value.
            show_items = [value_token]
        elif isinstance(error, ModelParameterDefaultValueError):
            # A referenced component doesn't have a default model parameter.
            target = data["token"].value
            msg = f"{msg} (hint: try '&{target}.[some parameter]')"
            # Mark only the value.
            show_items = [value_token]
        elif isinstance(error, ModelParameterSelfReferenceError):
            # Grab the production.
            item = production(param_node, graph)
            # Mark the key and the value.
            show_items = item.sorted_tokens

        super().__init__(msg, script, [show_items])


class KatParameterSelfReferenceException(Exception):
    """Indication that a parameter contains a self-reference that must be resolved
    later."""



import re
import pytest
from finesse.script.exceptions import KatScriptError


@pytest.fixture
def compiler(compiler):
    # Register an element.
    element = lambda a=None, b=None, c=None, d=None: None
    compiler.spec.register_element(
        "fake_element", {"setter": element, "getter": element}
    )
    return compiler


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_element myelement a=1 a=2",
            (
                "\nline 1: duplicate arguments with key 'a'\n"
                "-->1: fake_element myelement a=1 a=2\n"
                "                             ^   ^"
            ),
            id="duplicate keyword argument",
        ),
        pytest.param(
            "fake_element myelement a=1 b=2 a=3",
            (
                "\nline 1: duplicate arguments with key 'a'\n"
                "-->1: fake_element myelement a=1 b=2 a=3\n"
                "                             ^       ^"
            ),
            id="duplicate keyword argument, out of order",
        ),
        pytest.param(
            "fake_element myelement a=1 b=2 a=3 b=4",
            (
                "\nline 1: duplicate arguments with keys 'a' and 'b'\n"
                "-->1: fake_element myelement a=1 b=2 a=3 b=4\n"
                "                             ^   ^   ^   ^"
            ),
            id="multiple duplicate keyword arguments",
        ),
    ),
)
def test_duplicate_argument_error(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)



"""Containers for paths traversed through a configuration."""


class OpticalPath:
    """Represents a path traversing through optical connections of a :class:`.Model`
    instance.

    The underlying data stored by instances of this class are lists of two-element
    tuples containing optical nodes and the components that they connect into. This list
    is formatted as `[(from_node, to_comp)]` where `from_node` is an
    :class:`.OpticalNode` instance and `to_comp` can be any sub-class instance of
    :class:`.Connector`; `from_node` is then an input node to `to_comp`.

    A handle to the underlying list can be obtained through accessing the property
    :attr:`OpticalPath.data`. This is not required for iterating through the path
    entries however, as this class provides iterator access itself.

    Parameters
    ----------
    path : list
        A list of 2-tuples containing the path data; first element stores the
        :class:`.OpticalNode`, second element stores the component this node feeds into.
    """

    def __init__(self, path):
        self.__path = path

    def __str__(self):
        from tabulate import tabulate

        return tabulate(
            [(node.full_name, comp.name) for node, comp in self.__path],
            ["From optical node", "Into component"],
            tablefmt="fancy_grid",
            missingval="None",
        )

    def __iter__(self):
        return iter(self.__path)

    def __next__(self):
        return next(self.__path)

    @property
    def data(self):
        """A handle to the underlying path data.

        :getter: Returns the list of 2-tuples containing the path data.
        """
        return self.__path

    @property
    def nodes_only(self):
        """The path data with only the :class:`.OpticalNode` sequence.

        :getter: Returns a list of the sequence of traversed optical nodes.
        """
        return [pair[0] for pair in self.__path]

    @property
    def components_only(self):
        """The path data with only the component sequence.

        :getter: Returns a list of the sequence of traversed components.
        """
        return [pair[1] for pair in self.__path if pair[1] is not None]

    @property
    def spaces(self):
        """The spaces in the optical path.

        :getter: Yields the spaces in the optical path.
        """
        from finesse.components import Space

        for _, to_comp in self.__path:
            if isinstance(to_comp, Space):
                yield to_comp

    @property
    def length(self):
        """The length of the optical path.

        This returns the optical path length, i.e. the geometric length of each space scaled by its
        refractive index.

        :getter: Returns the total traversed length of the optical path (in metres).
        """
        return sum([space.L / space.nr for space in self.spaces])



"""Utility functions related to component objects."""


def refractive_index(port, symbolic=False):
    """Obtains the refractive index of the space attached to
    the port/node `port`.

    Parameters
    ----------
    port : :class:`.Port` or :class:`.Node`
        Port or node.

    Returns
    -------
    nr : float
        The refractive index of the space attached to the port / node.
        Returns unity if no space is present.
    """
    get_nr = lambda space, symbol: space.nr.ref if symbol else space.nr.eval()

    space = port.space
    if space is not None:
        nr = get_nr(space, symbolic)
    else:
        from finesse.components import Beamsplitter

        # If we're at a beamsplitter then get nr
        # of port on same surface
        if isinstance(port.component, Beamsplitter):
            adj_port = port.component.get_adjacent_port(port)
            space = adj_port.space
            if space is None:
                # FIXME (sjr) This should probably raise an exception but parsing
                #             (at least for legacy files) means that both spaces
                #             can be None initially somehow when symbolising
                #             Beamsplitter ABCD matrices
                nr = 1
            else:
                nr = get_nr(space, symbolic)
        else:
            nr = 1

    return nr



import re
import logging
import pytest
from finesse.script.exceptions import KatScriptError
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler):
    element = lambda a=None, b=None, c=None, d=None: None
    compiler.spec.register_element(
        "fake_element", {"setter": element, "getter": element}
    )
    cmd = lambda a, b: None
    compiler.spec.register_command(
        "fake_command", {"setter": cmd, "getter": cmd, "singular": True},
    )
    compiler.spec.constants = {"fake_constant": None}
    compiler.spec.keywords = {"fake_keyword"}

    return compiler


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "__undefined_element__ myelement 1 a=2",
            "\nline 1: unknown element '__undefined_element__'\n"
            "-->1: __undefined_element__ myelement 1 a=2\n"
            "      ^^^^^^^^^^^^^^^^^^^^^",
            id="element",
        ),
        pytest.param(
            "__undefined_function__(1, a=2)",
            "\nline 1: unknown function '__undefined_function__'\n"
            "-->1: __undefined_function__(1, a=2)\n"
            "      ^^^^^^^^^^^^^^^^^^^^^^",
            id="function",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_command(
                    1,
                    __undefined_function__(1, a=2)
                )
                """
            ),
            "\nline 3: unknown function '__undefined_function__'\n"
            "   2:     1,\n"
            "-->3:     __undefined_function__(1, a=2)\n"
            "          ^^^^^^^^^^^^^^^^^^^^^^\n"
            "   4: )",
            id="nested function",
        ),
    ),
)
def test_unknown_directive_error(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement
                fake_element myelement
                """
            ),
            (
                "\nlines 1-2: multiple elements with name 'myelement'\n"
                "-->1: fake_element myelement\n"
                "                   ^^^^^^^^^\n"
                "-->2: fake_element myelement\n"
                "                   ^^^^^^^^^"
            ),
            id="two elements with same name",
        ),
    ),
)
def test_duplicate_name_error(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            "fake_element fake_constant",
            (
                "\nline 1: constant 'fake_constant' cannot be used as an element name\n"
                "-->1: fake_element fake_constant\n"
                "                   ^^^^^^^^^^^^^"
            ),
            id="constant-as-a-name",
        ),
    ),
)
def test_invalid_name_error(compiler, script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        compiler.compile(script)


def test_name_warning(caplog, compiler):
    warning_text = (
        "element name 'fake_keyword' (line 1) is also the name of a keyword, which may "
        "lead to confusion"
    )
    with caplog.at_level(logging.WARNING, logger="finesse"):
        compiler.compile("fake_element fake_keyword")
        assert warning_text in caplog.text



"""Advanced beam tracing tests - i.e. interferometers with multiple cavities and configurations \
with several user set nodes and cavity objects."""

from collections import OrderedDict

from finesse.cymath.gaussbeam import transform_q
from finesse.gaussian import BeamParam
from finesse.script import parse_legacy


def test_michelson_trace():
    """Test beam tracing for a Michelson interferometer (no Fabry-Perot cavities) with a single \
    user-defined beam parameter at the starting laser node."""
    L0_q = -1.4 + 2.5j
    SCRIPT = r"""
    l L0 1 0 nL0
    s s1 0.5 nL0 nBS1

    bs BS 0.5 0.5 0 0 nBS1 nBS2 nBS3 nBS4

    s sy0 1.0 nBS2 nMY1
    m MY 0.5 0.5 0 nMY1 nMY2
    attr MY Rc 2.5

    s sx0 1.0 nBS3 nMX1
    m MX 0.5 0.5 0 nMX1 nMX2
    attr MX Rc 2.5
    """
    model = parse_legacy(SCRIPT)
    model.L0.p1.o.q = L0_q
    result = model.beam_trace().data_qx

    target = OrderedDict()
    # L0 p1
    target[model.L0.p1.o] = BeamParam(q=L0_q)
    target[model.L0.p1.i] = -target[model.L0.p1.o].conjugate()
    # BS p1
    target[model.BS.p1.i] = transform_q(
        model.L0.p1.o.space.ABCD(model.L0.p1.o, model.BS.p1.i),
        target[model.L0.p1.o],
        1,
        1,
    )
    target[model.BS.p1.o] = -target[model.BS.p1.i].conjugate()
    # BS p3
    target[model.BS.p3.o] = transform_q(
        model.BS.ABCD(model.BS.p1.i, model.BS.p3.o), target[model.BS.p1.i], 1, 1,
    )
    target[model.BS.p3.i] = -target[model.BS.p3.o].conjugate()
    # MX p1
    target[model.MX.p1.i] = transform_q(
        model.BS.p3.o.space.ABCD(model.BS.p3.o, model.MX.p1.i),
        target[model.BS.p3.o],
        1,
        1,
    )
    target[model.MX.p1.o] = -target[model.MX.p1.i].conjugate()
    # MX p2
    target[model.MX.p2.o] = transform_q(
        model.MX.ABCD(model.MX.p1.i, model.MX.p2.o), target[model.MX.p1.i], 1, 1,
    )
    target[model.MX.p2.i] = -target[model.MX.p2.o].conjugate()
    # BS p2
    target[model.BS.p2.o] = transform_q(
        model.BS.ABCD(model.BS.p1.i, model.BS.p2.o), target[model.BS.p1.i], 1, 1,
    )
    target[model.BS.p2.i] = -target[model.BS.p2.o].conjugate()
    # MY p1
    target[model.MY.p1.i] = transform_q(
        model.BS.p2.o.space.ABCD(model.BS.p2.o, model.MY.p1.i),
        target[model.BS.p2.o],
        1,
        1,
    )
    target[model.MY.p1.o] = -target[model.MY.p1.i].conjugate()
    # MY p2
    target[model.MY.p2.o] = transform_q(
        model.MY.ABCD(model.MY.p1.i, model.MY.p2.o), target[model.MY.p1.i], 1, 1,
    )
    target[model.MY.p2.i] = -target[model.MY.p2.o].conjugate()
    # BS p4
    target[model.BS.p4.o] = transform_q(
        model.BS.ABCD(model.BS.p2.i, model.BS.p4.o), target[model.BS.p2.i], 1, 1,
    )
    target[model.BS.p4.i] = -target[model.BS.p4.o].conjugate()

    for node, q in target.items():
        assert q == result[node]



"""Beam propagation tools for use outside of a simulation context.

The recommended function for most use-cases is
:func:`~finesse.tracing.tools.propagate_beam` (and
:func:`~finesse.tracing.tools.propagate_beam_astig` for astigmatic beam propagations)
which traces a beam through a specified path of a model. See :ref:`propagating_beams`
for details and examples.
"""

import numpy as np

from finesse.components import Cavity
from finesse.components.general import InteractionType
from finesse.gaussian import BeamParam
from finesse.paths import OpticalPath
from finesse.tracing.ctracer import TraceTree, get_last_left_branch
from finesse.solutions import (
    ABCDSolution,
    PropagationSolution,
    AstigmaticPropagationSolution,
)
from finesse.utilities import refractive_index

import finesse.tracing.cytools as cytools

import logging

LOGGER = logging.getLogger(__name__)


### Composite ABCD matrices ###


def compute_abcd(
    from_node=None,
    to_node=None,
    via_node=None,
    path=None,
    direction="x",
    symbolic=False,
    solution_name=None,
):
    """Computes the composite ABCD matrix through a given path.

    By setting the argument `symbolic` to true, this method will return a symbolic
    representation of the ABCD matrix rather than a numeric matrix.

    Parameters
    ----------
    from_node : :class:`.Node`
        Node to trace from.

    to_node : :class:`.Node`
        Node to trace to.

    via_node : :class:`.Node`, optional
        Optional node to trace via.

    path : :class:`.OpticalPath`, optional
        A pre-generated path to use (produced from a call to :meth:`.Model.path`).

    directon : str, optional; default: "x"
        Direction of ABCD matrix computation (can be 'x', for tangential plane,
        or 'y', for sagittal plane).

    symbolic : bool, optional; default: False
        Flag indicating whether to compute a symbolic ABCD matrix.

    Returns
    -------
    out : :class:`.ABCDSolution`
        ABCD matrix solution object between the specified nodes.
    """
    path = _make_path(from_node, to_node, via_node, path)
    # Make a tree from the forward path...
    t_initial = TraceTree.from_path(path.nodes_only)
    if t_initial is None:
        raise ValueError("Cannot compute ABCD matrix from a node to itself!")
    # ... then get the last branch as full ABCD is
    #     computed from multiplying each ABCD upwards
    #     through the full tree (i.e. correct multiplication order)
    t = get_last_left_branch(t_initial)

    t.node._model._update_symbolic_abcds()

    if symbolic:
        M = cytools.compute_symbolic_abcd(t, direction)
    else:
        M = cytools.compute_numeric_abcd(t, direction)

    fn = path.nodes_only[0]
    tn = path.nodes_only[-1]
    comp1 = fn.component
    comp2 = tn.component
    # Workaround for ABCD at single connector, reverting minus sign applied
    # due to co-ordinate system transformation on reflection
    if (
        comp1 is comp2
        and comp1.is_valid_coupling(fn, tn)
        and comp1.interaction_type(fn, tn) == InteractionType.REFLECTION
    ):
        M *= -1

    if solution_name is None:
        solution_name = f"ABCD_{fn.full_name}_{tn.full_name}_{direction}"
        if symbolic:
            solution_name += "_sym"

    return ABCDSolution(solution_name, M, direction, symbolic)


### Accumulated Gouy phases ###


def acc_gouy(
    from_node=None,
    to_node=None,
    via_node=None,
    path=None,
    q_in=None,
    direction="x",
    symbolic=False,
    deg=True,
    **kwargs,
):
    """Computes the accumulated Gouy phase along a specified path.

    By setting the argument `symbolic` to true, this method will return a symbolic
    representation of the accumulated Gouy phase rather than a number.

    If the argument `q_in` is not specified then this value will be determined
    from a call to :meth:`.Model.beam_trace`. Arguments to this beam trace call can be
    passed via the `kwargs` of this method.

    Parameters
    ----------
    from_node : :class:`.Node`
        Node to trace from.

    to_node : :class:`.Node`
        Node to trace to.

    via_node : :class:`.Node`, optional
        Optional node to trace via.

    path : :class:`.OpticalPath`, optional
        A pre-generated path to use.

    q_in : :class:`.BeamParam`, complex, optional
        Beam parameter to use at starting node. If not specified then
        this will be determined from a beam trace. Note that, if specified,
        this can also be a symbolic beam parameter.

    direction : str, optional; default: "x"
        Plane of computation (can be 'x', 'y' or `None`).

    symbolic : bool, optional; default: False
        Flag determining whether to return a symbolic representation.

    degrees : bool, optional; default: True
        Flag determining whether to convert return value from radians to degrees.
    """
    path = _make_path(from_node, to_node, via_node, path)
    t = TraceTree.from_path(path.nodes_only)
    if t is None:
        raise ValueError(
            "Cannot calculate accumulated Gouy phase from a node to itself!"
        )

    q_in = _make_input_q(q_in, t.node, direction, **kwargs)
    if q_in.wavelength != t.node._model.lambda0:
        LOGGER.warning(
            "In acc_gouy:\n"
            "    Wavelength of input beam parameter (%s m) not equal "
            "to wavelength of model associated with path (%s m).",
            q_in.wavelength,
            t.node._model.lambda0,
        )
    if q_in.symbolic and not symbolic:
        LOGGER.info(
            "In acc_gouy:\n"
            "    Specified q_in argument is symbolic, switching on "
            "symbolic Gouy phase accumulation."
        )
        symbolic = True

    if symbolic:
        return cytools.compute_symbolic_acc_gouy(t, q_in, direction, deg)

    return cytools.compute_numeric_acc_gouy(t, complex(q_in), direction, deg)


### Arbitrary beam propagations ###


def propagate_beam(
    from_node=None,
    to_node=None,
    via_node=None,
    path=None,
    q_in=None,
    direction="x",
    symbolic=False,
    solution_name=None,
    **kwargs,
):
    """Propagates a beam through a specified path, returning dictionaries of the beam
    parameter at each node and component.

    This method returns a :class:`.PropagationSolution` instance.

    See :ref:`propagating_beams` for details and examples on using this function.

    By setting the argument `symbolic` to true, this method will return symbolic
    representations of the beam parameters, ABCD matrices and accumulated Gouy phases.

    The argument `q_in` can be used to specify an arbitrary input beam parameter
    to be used at the starting node of the propagation. If not given then this
    will be determined from a call to :meth:`.Model.beam_trace`. Arguments to this
    beam trace call can be passed via the `kwargs` of this method.

    Parameters
    ----------
    from_node : :class:`.Node`
        Node to trace from.

    to_node : :class:`.Node`
        Node to trace to.

    via_node : :class:`.Node`, optional
        Optional node to trace via.

    path : :class:`.OpticalPath`, optional
        A pre-generated path to use.

    q_in : :class:`.BeamParam`, complex, optional
        Beam parameter to use at starting node. If not specified then
        this will be determined from a beam trace. Note that, if specified,
        this can also be a symbolic beam parameter.

    direction : str, optional; default: "x"
        Plane of computation (can be 'x', 'y' or `None`).

    symbolic : bool, optional; default: False
        Flag determining whether to return a symbolic representation.

    Returns
    -------
    ps : :class:`.PropagationSolution`
        A solution object for the propagation.
    """
    path = _make_path(from_node, to_node, via_node, path)
    t = TraceTree.from_path(path.nodes_only)
    if t is None:
        raise ValueError("Cannot propagate beam from a node to itself!")

    q_in = _make_input_q(q_in, t.node, direction, **kwargs)
    if q_in.wavelength != t.node._model.lambda0:
        LOGGER.warning(
            "In propagate_beam:\n"
            "    Wavelength of input beam parameter (%s m) not equal "
            "to wavelength of model associated with path (%s m).",
            q_in.wavelength,
            t.node._model.lambda0,
        )
    if q_in.symbolic and not symbolic:
        LOGGER.info(
            "In propagate_beam:\n"
            "    Specified q_in argument is symbolic, switching on "
            "symbolic propagation."
        )
        symbolic = True

    if symbolic:
        node_info, comp_info = cytools.propagate_beam_symbolic(t, q_in, direction)
    else:
        node_info, comp_info = cytools.propagate_beam_numeric(t, q_in, direction)

    if solution_name is None:
        fn = path.nodes_only[0]
        tn = path.nodes_only[-1]

        solution_name = f"Propagation_{fn.full_name}_{tn.full_name}_{direction}"
        if symbolic:
            solution_name += "_sym"

    return PropagationSolution(solution_name, node_info, comp_info, symbolic)


def propagate_beam_astig(
    from_node=None,
    to_node=None,
    via_node=None,
    path=None,
    qx_in=None,
    qy_in=None,
    symbolic=False,
    solution_name=None,
    **kwargs,
):
    """Propagates the beam through a specified path over both the tangential and
    sagittal planes.

    Internally this calls :func:`~finesse.tracing.tools.propagate_beam` twice - for both
    the tangential and sagittal planes - and returns a solution object which stores the
    returns of these as properties.

    Parameters
    ----------
    from_node : :class:`.Node`
        Node to trace from.

    to_node : :class:`.Node`
        Node to trace to.

    via_node : :class:`.Node`, optional
        Optional node to trace via.

    path : :class:`.OpticalPath`, optional
        A pre-generated path to use.

    qx_in : :class:`.BeamParam`, complex, optional
        Beam parameter, in the tangential plane, to use at starting node. If not
        specified then this will be determined from a beam trace. Note that,
        if specified, this can also be a symbolic beam parameter.

    qy_in : :class:`.BeamParam`, complex, optional
        Beam parameter, in the sagittal plane, to use at starting node. If not
        specified then this will be determined from a beam trace. Note that,
        if specified, this can also be a symbolic beam parameter.

    symbolic : bool, optional; default: False
        Flag determining whether to return a symbolic representation.

    Returns
    -------
    astig_sol : :class:`.AstigmaticPropagationSolution`
        A solution object consisting of the propagation solutions for both planes
        and methods for accessing the per-plane beam parameters and overlaps.
    """
    path = _make_path(from_node, to_node, via_node, path)
    ps_x = propagate_beam(
        path=path, q_in=qx_in, direction="x", symbolic=symbolic, **kwargs
    )
    ps_y = propagate_beam(
        path=path, q_in=qy_in, direction="y", symbolic=symbolic, **kwargs
    )

    if solution_name is None:
        fn = path.nodes_only[0]
        tn = path.nodes_only[-1]

        solution_name = f"AstigProp_{fn.full_name}_{tn.full_name}"
        if symbolic:
            solution_name += "_sym"

    return AstigmaticPropagationSolution(solution_name, ps_x, ps_y)


### Computing mode mismatches ###


def compute_cavity_mismatches(model, cav1=None, cav2=None):
    """Computes the mismatch parameter (see :meth:`.BeamParam.mismatch` for the
    equation) between cavities of the model.

    If either / both of `cav1`, `cav2` are not specified then these will be set to
    all the cavities of the model. This means that the default behaviour of this method
    (specifying no args) is to compute mismatches between each cavity in the model.

    If either of each cavity in a coupling is unstable then the mismatch values
    between these will be given as ``np.nan``.

    Parameters
    ----------
    cav1 : :class:`.Cavity`, str, optional; default: None
        A single cavity object (or its name). Defaults to None such that
        all cavities are used.

    cav2 : :class:`.Cavity`, str, optional; default: None
        A single cavity object (or its name). Defaults to None such that
        all cavities are used.

    Returns
    -------
    mmx : float or dict
        If both `cav1` and `cav2` were specified then this will be a single number
        giving the mismatch between these cavities in the tangential plane.

        Otherwise, mmx is a dictionary of ``(c1, c2): mm_x`` mappings, where `c1`
        and `c2` are the cavity names and `mm_x` is the mismatch between any two
        cavities in the tangential plane.

    mmy : float or dict
        If both `cav1` and `cav2` were specified then this will be a single number
        giving the mismatch between these cavities in the sagittal plane.

        Otherwise, mmy is a dictionary of ``(c1, c2): mm_y`` mappings, where `c1`
        and `c2` are the cavity names and `mm_y` is the mismatch between any two
        cavities in the sagittal plane.
    """
    if cav1 is None:
        cavs_outer = model.cavities
    else:
        if isinstance(cav1, str):
            _c1 = model.elements.get(cav1)
        else:
            _c1 = cav1
        if not isinstance(_c1, Cavity):
            raise ValueError(f"Invalid argument type/name for cav1: {cav1}")

        cavs_outer = [_c1]

    if cav2 is None:
        cavs_inner = model.cavities
    else:
        if isinstance(cav2, str):
            _c2 = model.elements.get(cav2)
        else:
            _c2 = cav2
        if not isinstance(_c2, Cavity):
            raise ValueError(f"Invalid argument type/name for cav2: {cav2}")

        cavs_inner = [_c2]

    mm_x = {}
    mm_y = {}
    for c1 in cavs_outer:
        if not c1.is_stable:
            for c2 in cavs_inner:
                mm_x[(c1.name, c2.name)] = mm_x[(c2.name, c1.name)] = np.nan
                mm_y[(c1.name, c2.name)] = mm_y[(c2.name, c1.name)] = np.nan
        else:
            trace = model.beam_trace(store=False, enable_only=c1)
            for c2 in cavs_inner:
                if not c2.is_stable:
                    mm_x[(c1.name, c2.name)] = mm_x[(c2.name, c1.name)] = np.nan
                    mm_y[(c1.name, c2.name)] = mm_y[(c2.name, c1.name)] = np.nan
                elif c1 is c2:
                    mm_x[(c1.name, c2.name)] = mm_x[(c2.name, c1.name)] = 0
                    mm_y[(c1.name, c2.name)] = mm_y[(c2.name, c1.name)] = 0
                else:
                    q1x, q1y = trace[c2.source]
                    q2x, q2y = c2.qx, c2.qy

                    mmx = BeamParam.mismatch(q1x, q2x)
                    mmy = BeamParam.mismatch(q1y, q2y)

                    mm_x[(c1.name, c2.name)] = mm_x[(c2.name, c1.name)] = mmx
                    mm_y[(c1.name, c2.name)] = mm_y[(c2.name, c1.name)] = mmy

    if cav1 is not None and cav2 is not None:
        return list(mm_x.values())[0], list(mm_y.values())[0]

    return mm_x, mm_y


### Convenience functions ###


def _make_path(from_node, to_node, via_node, path):
    if path is not None:
        if from_node is not None or to_node is not None or via_node is not None:
            raise ValueError(
                "Cannot specify both path and any of " "(from_node, to_node, via_node)."
            )

        if not isinstance(path, OpticalPath):
            raise TypeError(
                "Expected path to be of type OpticalPath (i.e. return "
                "value of a Model.path call), but got an object of "
                f"type: {type(path)}"
            )
    else:
        if from_node is None or to_node is None:
            raise ValueError("One of: path OR (from_node, to_node) must be specified.")

        model = from_node._model
        if to_node._model is not model:
            raise ValueError(
                f"{from_node.full_name} and {to_node.full_name} are from "
                "different models."
            )
        path = model.path(from_node, to_node, via_node=via_node)

    return path


def _make_input_q(q_in, node, direction, **kwargs):
    model = node._model
    if q_in is None:
        trace = model.beam_trace(store=False, **kwargs)

        qx_in, qy_in = trace[node]
        if direction == "x":
            q_in = qx_in
        else:
            q_in = qy_in
    else:
        model._update_symbolic_abcds()

        if not isinstance(q_in, BeamParam):
            q_in = BeamParam(
                q=q_in, wavelength=model.lambda0, nr=refractive_index(node),
            )

    return q_in



import click
import tabulate
from .util import (
    table_format_argument,
    verbose_option,
    quiet_option,
    debug_option,
    log_exclude_option,
    KatState,
)
from ..config import config_instance


def config_edit(ctx, option, value):
    """Open the user configuration in the system default editor."""
    if not value or ctx.resilient_parsing:  # Keep this check as-is!
        return

    state = ctx.ensure_object(KatState)

    state.print("Opening user configuration in system editor...", fg="green", bold=True)
    click.edit(filename=config_instance().user_config_path(), extension=".ini")
    state.exit()


def config_reset(ctx, option, value):
    """Reset the user configuration to the default."""
    if not value or ctx.resilient_parsing:  # Keep this check as-is!
        return

    state = ctx.ensure_object(KatState)

    click.confirm(
        f"Are you sure you wish to reset the user configuration at "
        f"{config_instance().user_config_path()}?",
        abort=True,
    )

    try:
        config_instance().write_user_config(force=True)
    except Exception as e:
        state.print_error(e, title="ERROR:")

    state.print(
        "Successfully reset user configuration.", fg="green", bold=True, exit_=True
    )


@click.command()
@click.option(
    "--paths/--no-paths",
    is_flag=True,
    default=True,
    show_default=True,
    help="Show the paths used to build the configuration.",
)
@click.option(
    "--dump/--no-dump",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show the current configuration in a table.",
)
@click.option(
    "--edit",
    is_flag=True,
    callback=config_edit,
    expose_value=False,
    is_eager=True,
    help="Open the user configuration in the system default editor.",
)
@click.option(
    "--reset",
    is_flag=True,
    callback=config_reset,
    expose_value=False,
    is_eager=True,
    help="Reset the user configuration to the default.",
)
@table_format_argument
@verbose_option
@quiet_option
@debug_option
@log_exclude_option
@click.pass_context
def config(ctx, paths, dump, table_format):
    """Configuration information."""
    state = ctx.ensure_object(KatState)
    config = config_instance()

    if paths:
        state.print(
            "Configuration paths (1 = highest priority; ✓ = found, ✗ = not found):",
            fg="green",
            bold=True,
        )

        configs = config.user_config_paths().items()

        for priority, (_, path) in enumerate(reversed(configs), start=1):
            mark = "✓" if path.exists() else "✗"
            pathstr = click.format_filename(str(path))
            state.print(f"{mark} {priority}: {pathstr}", indent=1)

    if dump:
        rows = []
        for section in config.sections():
            _section = click.style(section, fg="green", bold=True)
            for option, value in config.items(section):
                option = click.style(option, fg="blue", bold=True)
                rows.append((_section, option, value))

        state.print(tabulate.tabulate(rows, tablefmt=table_format))


if __name__ == "__main__":
    config()



""""""
import finesse
import pytest


@pytest.fixture
def model():
    IFO = finesse.Model()
    IFO.parse(
        """
    # Radii of curvature of arm cavity mirrors
    var XROC1 -1934
    var XROC2 2245
    var YROC1 -1934
    var YROC2 2245

    l L0 P=1
    link(L0, BS)
    bs BS

    s sy BS.p2 ITMY.p1 L=10
    m ITMY R=0.99 T=0.01 Rc=[&YROC1, &YROC1]
    s LY ITMY.p2 ETMY.p1 L=4k
    m ETMY R=0.99 T=0.01 Rc=[&YROC2, &YROC2]

    s sx BS.p3 ITMX.p1 L=10
    m ITMX R=0.99 T=0.01 Rc=[&XROC1, &XROC1]
    s LX ITMX.p2 ETMX.p1 L=4k
    m ETMX R=0.99 T=0.01 Rc=[&XROC2, &XROC2]

    cav cavXARM ITMX.p2
    cav cavYARM ITMY.p2

    modes(maxtem=2)

    knmd K22_itmx ITMX 22

    # This will detect mode mismatch from BS.p2.i -> BS.p4.o
    mmd mm_bs_x BS.p2.i BS.p4.o direction=x
    mmd mm_bs_y BS.p2.i BS.p4.o direction=y
    """
    )
    return IFO


def test_xaxis(model):
    model.parse("xaxis(XROC1, lin, -1900, -2000, 1)")
    model.run()


def test_noxaxis(model):
    model.run()



"""Kat script utility functions."""

from itertools import groupby
from collections import defaultdict


def index_ranges(intlist):
    """Create a sequence of range strings from a sequence of integers.

    Based on https://stackoverflow.com/a/9471386/2251982.

    Parameters
    ----------
    intlist : sequence
        Integers to convert to range strings.

    Yields
    ------
    str
        The next range string.

    Examples
    --------
    >>> list(index_ranges([1, 2, 4, 5, 6, 7, 9]))
    ["1-2", "4-7", "9"]
    """

    def sub(x):
        return x[1] - x[0]

    for k, iterable in groupby(enumerate(sorted(intlist)), sub):
        indices = list(iterable)
        if len(indices) == 1:
            yield str(indices[0][1])
        else:
            yield f"{indices[0][1]}-{indices[-1][1]}"


def scriptsorted(items, reverse=False):
    """Sort container items by their script position.

    Parameters
    ----------
    items : sequence of :class:`.TokenContainer`
        The container items to sort.

    reverse : bool, optional
        Request the result in descending instead of ascending order.

    Returns
    -------
    list of :class:`.TokenContainer`
        Container items sorted by line number then column.
    """
    return sorted(items, key=lambda item: item.start, reverse=reverse)


def duplicates(items, key=None):
    """Get duplicate keys and values in the 1D sequence `items`.

    Parameters
    ----------
    items : sequence
        The sequence to find duplicates in.

    key : callable, optional
        The key function to use for comparisons. If not specified, defaults to the
        identity function.

    Returns
    -------
    list
        The duplicates in `items`. This is a sequence of tuples containing the result
        of the key function for each entry of `items`, where at least two such keys
        exist, and the original items that matched that `key`.

    Examples
    --------
    >>> [k for k, _ in duplicates("AAAABBCDAAB")]
    ["A", "B", "C", "D"]

    >>> [list(g) for _, g in duplicates("AAAABBCDAAB")]
    [["A", "A", "A", "A", "A", "A"], ["B", "B", "B"], ["C"], ["D"]]
    """
    if key is None:
        key = lambda item: item
    groups = defaultdict(list)
    for item in items:
        groups[key(item)].append(item)
    return [(k, v) for k, v in groups.items() if len(v) > 1]


def merge_attributes(attr1, attr2):
    """Perform a deep attribute dictionary merge.

    This merges list and tuples in `attr1` and `attr2` into one dict. Values that are
    not lists or tuples result in an error. Items in `attr1` appear before those in
    `attr2` where they share the same key.

    Parameters
    ----------
    attr1, attr2 : dict
        The attribute dictionaries to merge.

    Returns
    -------
    dict
        The merged attribute dictionary.

    Raises
    ------
    ValueError
        If a dict value in `attr1` or `attr2` is not a list or tuple.
    """
    out = {}
    secondkeys = set(attr2)

    # Add keys in attr1 and optionally attr2.
    for key, value in attr1.items():
        if isinstance(value, list):
            value = [*value, *attr2.get(key, [])]
        elif isinstance(value, tuple):
            value = (*value, *attr2.get(key, ()))
        elif key in attr2:
            raise ValueError(
                f"don't know how to merge {type(value)} and {type(attr2[key])}"
            )

        out[key] = value
        secondkeys.discard(key)

    # Add keys only in attr2.
    for key in secondkeys:
        out[key] = attr2[key]

    return out



"""
Objects for connecting and registering connections between components.
"""

import enum
import logging
import weakref

import finesse

from collections import OrderedDict
from copy import deepcopy
from finesse.exceptions import ComponentNotConnected
from finesse.freeze import Freezable
from .gauss import Gauss
from finesse.utilities import check_name, is_iterable


LOGGER = logging.getLogger(__name__)


@enum.unique
class NodeType(enum.Enum):
    """
    Enum describing the physical connection type of a :class:`.Node`
    """

    OPTICAL = 0
    ELECTRICAL = 1
    MECHANICAL = 2


@enum.unique
class NodeDirection(enum.Enum):
    """
    Enum describing the direction that information at a :class:`.Node`
    flows.

    This is largely a description to help understand how external information
    flows in and out of a component. Inside a component all nodes will
    couple to one another in more complex ways.

    Input nodes are those going into a component, whereas output
    describe those leaving. For example incident and reflected light fields.

    Bidrectional takes information either direction. For example a mechanical
    degree of freedom, external forces can be applied to it, or
    its motion can be coupled to some external system.
    """

    INPUT = 0
    OUTPUT = 1
    BIDIRECTIONAL = 2


class Port(Freezable):
    """A collection of all the nodes at a specific point/surface of a component.

    Parameters
    ----------
    name : str
        Name of newly created node.

    component : Sub-class of :class:`.Connector`
        The component that this node belongs to.

    node_type : :class:`.NodeType`
        Physical node type.
    """

    def __init__(self, name, component, node_type):
        self._unfreeze()
        self.__component = weakref.ref(component)
        self.__name = check_name(name)
        self.__full_name = "{}.{}".format(component.name, name)
        self.__type = node_type
        self.__nodes = OrderedDict()
        self.__enabled = True
        self._freeze()

    def __deepcopy__(self, memo):
        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))
        # Manually update the weakrefs to be correct
        new.__component = weakref.ref(memo[id(self.component)])
        return new

    def _add_node(self, name, direction=None, node=None, **kwargs):
        """
        Adds a new node to this port. Once called the new node
        can be access with:

            obj.port.name

        Parameters
        ----------
        name : str
            Name of the new node
        dof_parameter : ModelParameter
            A model parameter to be associated with this node
        """
        if name in self.__nodes:
            raise Exception("Node already added to this component")

        if node is None:
            if self.__type == NodeType.OPTICAL:
                if direction is None:
                    raise Exception(
                        "Node direction must be specified for optical nodes"
                    )
                node = OpticalNode(name, self, direction, **kwargs)
            elif self.__type == NodeType.MECHANICAL:
                node = MechanicalNode(name, self, direction, **kwargs)
            elif self.__type == NodeType.ELECTRICAL:
                node = ElectricalNode(name, self, direction, **kwargs)
            else:
                raise Exception("Unexpected node type")

        elif len(kwargs) != 0:
            raise Exception("Cannot use kwargs when node is not None")

        assert node is not None
        
        if node.type != self.type:
            raise Exception("Node and port type must be the same")

        self.__nodes[name] = node
        # Update Elements node dict so it doesn't have to keep checking
        # it's ports about what nodes it has
        self.__component()._Connector__nodes[node.full_name] = node
        self._unfreeze()
        assert not hasattr(self, name)
        setattr(self, name, node)
        self._freeze()

    def __repr__(self):
        return f"<Port {self.component.name}.{self.name} Type={self.__type} @ {hex(id(self))}>"

    @property
    def full_name(self):
        """
        :getter: Returns a full name of the port: {component name}.{port name}
        """
        return self.__full_name

    @property
    def enabled(self):
        return self.__enabled

    @property
    def type(self):
        """:class:`.NodeType` of the port object.

        :getter: Returns the node-type of the port (read-only).
        """
        return self.__type

    @property
    def name(self):
        """Name of the port object.

        :getter: Returns the name of the port (read-only).
        """
        return self.__name

    @property
    def _model(self):
        return self.component._model

    @property
    def component(self):
        """The component which has ownership of this port.

        :getter: Returns the component that this port belongs to (read-only).
        """
        return self.__component()

    @property
    def is_connected(self):
        """Flag indicating whether the port is attached to another component.

        :getter: Returns true if this port is attached to another component (read-only).
        """
        first = self.nodes[0].connection
        if first is None:
            return False
        else:
            return all(first is n.connection for n in self.nodes)

    @property
    def attached_to(self):
        """Component that the port is attached to.

        :getter: Returns the component this port is attached to, or returns None
                 if no such connected component exists (read-only).
        """
        if self.type == NodeType.OPTICAL:
            spaces = [_.space for _ in self.nodes]
            if all([spaces[0] == _ for _ in spaces]):
                return spaces[0]
            else:
                raise Exception(
                    "Nodes are somehow connected to different "
                    "spaces which should not happen"
                )
        else:
            raise NotImplementedError()

    @property
    def space(self):
        """Space that the port is attached to. Equivalent to :attr:`Port.attached_to`.

        :getter: Returns the space that this port is attached to (read-only).
        """
        if self.type != NodeType.OPTICAL:
            raise Exception("Port type is not optical, cannot retrieve attached space.")
        return self.attached_to

    @property
    def nodes(self):
        """Nodes associated with the port.

        :getter: Returns a tuple of the associated nodes at this port (read-only).
        """
        return tuple(self.__nodes.values())


class Node:
    """Represents a specific connection at a component.

    Mathematically a node represents a single equation in the interferometer matrix.

    A node can only be owned by a single component instance - with weak references stored by the
    connected components.

    Parameters
    ----------
    name : str
        Name of newly created node.

    component : :class:`.Port`
        The port that this node belongs to.

    node_type : :class:`.NodeType`
        Physical node type.
    """

    def __init__(self, name, port, node_type, direction):
        self.__port = weakref.ref(port)
        # self.__component = weakref.ref(port.component)
        self.__name = check_name(name)
        self.__type = node_type
        self.__direction = direction
        self.__full_name = "{}.{}.{}".format(port.component.name, port.name, name)
        self.__port_name = "{}.{}".format(port.name, name)
        self.__tag_name = None
        self.__connection = None
        self.used_in_detector_output = []
        self.has_signal_injection = False

    def __deepcopy__(self, memo):
        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))
        id_port = id(self.port)
        # Manually update the weakrefs to be correct
        if id_port in memo:
            new.__port = weakref.ref(memo[id_port])
        else:
            # We need to update this reference later on
            # This will be called when the port property
            # is accessed. When this happens we'll peak back
            # at the memo once it has been filled and get
            # the new port reference. After this the refcount
            # for this function should goto zero and be garbage
            # collected
            def update_later():
                new.__port = weakref.ref(memo[id_port])

            new.__port = update_later  # just in case something calls
            # this weakref in the meantime
            memo[id(self._model)].after_deepcopy.append(update_later)

        # new.__component = weakref.ref(memo[id(self.component)])
        return new

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.full_name} @ {hex(id(self))}>"

    @property
    def is_input(self):
        """Flag indicating whether this node is an input to
        the associated component.

        :getter: Returns `True` if the field at this node
                 goes into `self.component` (read-only).
        """
        return self.direction == NodeDirection.INPUT

    @property
    def full_name(self):
        """Full name

        :getter: Returns a full name of the node: {component name}.{port name}.{node name}
        """
        return self.__full_name

    @property
    def port_name(self):
        """Port name

        :getter: Returns a shortened name of the node: {port name}.{node name}
        """
        return self.__port_name

    @property
    def type(self):
        """:class:`.NodeType` of the node object.

        :getter: Returns the node-type of the node (read-only).
        """
        return self.__type

    @property
    def direction(self):
        """
        :class:`.NodeDirection` of this node. This is largely a
        description to help understand how external information flow
        in and out of a component. Inside a component all nodes will
        couple to one another in more complex ways.

        Input nodes are those going into a component, whereas output
        describe those leaving. For example incident and reflected light fields.

        Bidrectional takes information either direction. For example a mechanical
        degree of freedom, external forces can be applied to it, or
        its motion can be coupled to some external system.

        :getter: Returns the directionality of the node (read-only).
        """
        return self.__direction

    @property
    def port(self):
        """:class:`.Port` this node is attached to.

        :getter: Returns the port of this node (read-only).
        """
        return self.__port()

    @property
    def name(self):
        """Name of the node object.

        :getter: Returns the name of the node (read-only).
        """
        return self.__name

    @property
    def tag(self):
        """Tagged name of the node object.

        :getter: Returns the tagged (user-defined) name of the node (read-only).
        """
        return self.__tag_name

    def _set_tag(self, tag):
        self.__tag_name = tag

    @property
    def _model(self):
        return self.component._model

    @property
    def component(self):
        """The component which has ownership of this node.

        :getter: Returns the component that this node belongs to (read-only).
        """
        return self.port.component

    def is_neighbour(self, node):
        """Checks if `node` is a connected by an edge to this node.

        Parameters
        ----------
        node : :class:`.Node`
            Node with which to check connection.

        Returns
        -------
        flag : bool
            True if `node` is connected to this node, False otherwise.
        """
        # if not associated with a model yet, check registered connections directly
        try:
            a = self._model.network.has_node(self.full_name)
            b = node.full_name in self._model.network.neighbors(self.full_name)
            return a and b
        except ComponentNotConnected:
            return (
                self.full_name,
                node.full_name,
            ) in self.component._registered_connections.values()

        # if not self.has_model():
        #     return (self, node) in self.__component()._registered_connections.values()
        # return (self._model.network.has_node(self) and
        #         node in self._model.network.neighbors(self))

    @property
    def connection(self):
        """A reference to the :class:`.Space`, :class:`.Wire`, or :class:`.Joint`
        object attached to this node.

        :getter: Returns a reference to the :class:`.Space`, :class:`.Wire`, or :class:`.Joint`
                object attached to the node (read-only).
        """
        # if self.__connection is not None:
        #    return self.__connection()

        try:
            if self.direction == NodeDirection.INPUT:
                edges = self._model.network.in_edges(self.full_name)
            elif self.direction == NodeDirection.OUTPUT:
                edges = self._model.network.out_edges(self.full_name)
            elif self.direction == NodeDirection.BIDIRECTIONAL:
                raise NotImplementedError()
            else:
                raise NotImplementedError(f"{self}, {self.direction}")

            for edge in edges:
                edge_data = self._model.network.get_edge_data(*edge)
                owner = edge_data["owner"]()
                if isinstance(
                    owner,
                    (
                        finesse.components.Space,
                        finesse.components.Wire,
                        finesse.components.Joint,
                        finesse.components.DegreeOfFreedom,
                    ),
                ):
                    # self.__connection = edge_data["owner"]
                    return owner
        except ComponentNotConnected:
            pass

        return None


class MechanicalNode(Node):
    """Represents a specific mechanical degree of freedom of a component.

    Parameters
    ----------
    name : str
        Name of the mechanical motion.

    port : :class:`.Port`
        The port that this node belongs to.

    num_frequencies : int
        Number of mechanical frequencies to model
    """

    def __init__(self, name, port, direction):
        super().__init__(name, port, NodeType.MECHANICAL, direction)
        self.__frequencies = None
        self.__num_frequencies = 1

    @property
    def num_frequencies(self):
        return self.__num_frequencies

    @property
    def frequencies(self):
        if self.__frequencies is None:
            return (self.component._model.fsig.f.ref,)
        else:
            return self.__frequencies

    @frequencies.setter
    def frequencies(self, value):
        self.__frequencies = value
        self.__num_frequencies = len(value)


class ElectricalNode(Node):
    """Represents a specific electrical degree of freedom of a component.

    Parameters
    ----------
    name : str
        Name of the mechanical motion.

    port : :class:`.Port`
        The port that this node belongs to.
    """

    def __init__(self, name, port, direction):
        super().__init__(name, port, NodeType.ELECTRICAL, direction)
        self.__frequencies = None
        self.__num_frequencies = 1

    @property
    def num_frequencies(self):
        return self.__num_frequencies

    @property
    def frequencies(self):
        if self.__frequencies is None:
            return (self.component._model.fsig.f.ref,)
        else:
            return self.__frequencies

    @frequencies.setter
    def frequencies(self, value):
        self.__frequencies = value
        self.__num_frequencies = len(value)


class OpticalNode(Node):
    """Represents a specific optical port connection at a component.

    OpticalNodes also have additional physical properties such as the beam parameter (of type
    :class:`.BeamParam`) at the nodes' position within the interferometer.

    Parameters
    ----------
    name : str
        Name of the optical node.
    port : :class:`.Port`
        The port that this node belongs to.
    direction : :class:`.NodeDirection`
        True if the field at this node is going into the component.
    """

    def __init__(self, name, port, direction):
        super().__init__(name, port, NodeType.OPTICAL, direction)

        self.__space = None

    def __deepcopy__(self, memo):
        new = super().__deepcopy__(memo)
        # Manually update the weakrefs to be correct
        if self.__space is not None:
            id_space = id(self.__space())
            if id_space in memo:
                new.__space = weakref.ref(memo[id_space])
            else:
                # We need to update this reference later on
                # This will be called when the port property
                # is accessed. When this happens we'll peak back
                # at the memo once it has been filled and get
                # the new port reference. After this the refcount
                # for this function should goto zero and be garbage
                # collected
                def update_later():
                    new.__space = weakref.ref(memo[id_space])

                new.__space = update_later  # just in case something calls
                # this weakref in the meantime
                memo[id(self._model)].after_deepcopy.append(update_later)

        return new

    @property
    def opposite(self):
        """The opposite direction node.

        :getter: Returns the opposite direction node to this one.
        """
        return getattr(self.port, "o" if self.is_input else "i")

    @property
    def q(self):
        """Beam parameter value at this node.

        :getter: Returns the beam parameter at this node. If the beam
                 parameters in the tangential and sagittal planes are
                 different then it returns a tuple of the two parameters.
        :setter: Sets the beam parameter at this node. If the argument
                 provided is a 2-tuple then the parameter is set astigmatically
                 for the node.
        """
        return self.qx, self.qy

    @q.setter
    def q(self, value):
        if is_iterable(value):  # both qx and qy specified
            self.qx = value[0]
            self.qy = value[1]
        else:  # only one q specified
            if self in self._model.gauss_commands:
                self._model.update_gauss(self, qx=value, qy=value)

            else:
                node_name = self.full_name.replace(".", "_")
                name = f"g{node_name}"
                self._model.add(Gauss(name, self, qx=value, qy=value))

    @property
    def qx(self):
        """Beam parameter value in the tangential plane.

        :getter: Returns the beam parameter at the node in the
                 tangential plane.
        :setter: Sets the beam parameter at the node in the
                 tangential plane.
        """
        trace = self._model.last_trace
        if trace is None:
            gauss = self._model.gauss_commands.get(self)
            # TODO (sjr) Also check in model.cavities for this node as source
            if gauss is None:
                raise RuntimeError(
                    f"No stored beam trace yet performed on model and {self.full_name} "
                    "does not have correspond to any Gauss or Cavity object - unable to "
                    "access any beam parameter at this node."
                )

            LOGGER.warning(
                "No beam trace solution yet stored in the model. Returning "
                "qx based on Gauss object entry."
            )
            qx = gauss.qx
        else:
            qx, _ = trace.get(self, (None, None))
            if qx is None:
                raise RuntimeError(
                    f"Bug encountered! Could not find entry for {self.full_name} "
                    "in last stored beam trace solution of the model."
                )

        return qx

    @qx.setter
    def qx(self, value):
        if self in self._model.gauss_commands:
            self._model.update_gauss(self, qx=value)

        else:
            node_name = self.full_name.replace(".", "_")
            name = f"g{node_name}"
            self._model.add(Gauss(name, self, q=value))

    @property
    def qy(self):
        """Beam parameter value in the sagittal plane.

        :getter: Returns the beam parameter at the node in the
                 sagittal plane.
        :setter: Sets the beam parameter at the node in the
                 sagittal plane.
        """
        trace = self._model.last_trace
        if trace is None:
            gauss = self._model.gauss_commands.get(self)
            # TODO (sjr) Also check in model.cavities for this node as source
            if gauss is None:
                raise RuntimeError(
                    f"No stored beam trace yet performed on model and {self.full_name} "
                    "does not have correspond to any Gauss or Cavity object - unable to "
                    "access any beam parameter at this node."
                )

            LOGGER.warning(
                "No beam trace solution yet stored in the model. Returning "
                "qy based on Gauss object entry."
            )
            qy = gauss.qy
        else:
            _, qy = trace.get(self, (None, None))
            if qy is None:
                raise RuntimeError(
                    f"Bug encountered! Could not find entry for {self.full_name} "
                    "in last stored beam trace solution of the model."
                )

        return qy

    @qy.setter
    def qy(self, value):
        if self in self._model.gauss_commands:
            self._model.update_gauss(self, qy=value)

        else:
            node_name = self.full_name.replace(".", "_")
            name = f"g{node_name}"
            self._model.add(Gauss(name, self, q=value))

    @property
    def space(self):
        """A reference to the :class:`.Space`
        object attached to this node.

        :getter: Returns a reference to the :class:`.Space`
                 object attached to the node (read-only).
        """
        if self.__space is not None:
            return self.__space()

        try:
            if self.is_input:
                edges = self._model.network.in_edges(self.full_name)
            else:
                edges = self._model.network.out_edges(self.full_name)

            for edge in edges:
                edge_data = self._model.network.get_edge_data(*edge)
                owner = edge_data["owner"]()

                if isinstance(owner, finesse.components.Space):
                    self.__space = weakref.ref(owner)
                    return owner
        except ComponentNotConnected:
            pass

        return None



"""Tests on modal properties of cavity detector."""

import numpy as np
from numpy.testing import assert_allclose
import pytest
from finesse import Model
from finesse.analysis import noxaxis
from finesse import constants


@pytest.fixture(scope="module")
def fp_cavity_script():
    return """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM Rc=-2
    s sc ITM.p2 ETM.p1 L=1
    m ETM Rc=2

    cav FP ITM.p2
    """


@pytest.mark.parametrize(
    "L, Rc", ((1, 1), (1, 2), (1, np.inf), (4e3, 2090), (10e3, 5580), (2, 0.2)),
)
def test_symm_fp_cav_stability(model: Model, fp_cavity_script, L, Rc):
    """Test detecting of g-factor for a symmetric Fabry-Perot cavity, using analytics to
    compare results."""
    model.parse(fp_cavity_script)
    # Add a mode matched gauss at laser so that unstable cavity
    # values still run the file
    model.add_matched_gauss(model.L0.p1.o)
    model.parse("cp gx FP g x")
    model.parse("cp gy FP g y")
    model.ITM.Rc = -1 * Rc
    model.ETM.Rc = Rc
    model.elements["sc"].L = L

    gcav_analytic = lambda L, Rc: (1 - L / Rc) ** 2

    out = noxaxis(model)

    assert abs(out["gx"] - out["gy"]) == pytest.approx(0)
    assert out["gx"] == pytest.approx(gcav_analytic(L, Rc))


@pytest.mark.parametrize(
    "L, Rc1, Rc2",
    (
        (1, 1, 0.9),
        (1, 1.9, 2.1),
        (1, np.inf, 2),
        (3994, 1934, 2245),
        (10e3, 5500, 5700),
        (10, 12, np.inf),
    ),
)
def test_asymm_fp_cav_stability(model: Model, fp_cavity_script, L, Rc1, Rc2):
    """Test detecting of g-factor for an asymmetric Fabry-Perot cavity, using analytics
    to compare results."""
    model.parse(fp_cavity_script)
    # Add a mode matched gauss at laser so that unstable cavity
    # values still run the file
    model.add_matched_gauss(model.L0.p1.o)
    model.parse("cp gx FP g x")
    model.parse("cp gy FP g y")
    model.ITM.Rc = -1 * Rc1
    model.ETM.Rc = Rc2
    model.elements["sc"].L = L

    gcav_analytic = lambda L, Rc1, Rc2: (1 - L / Rc1) * (1 - L / Rc2)

    out = noxaxis(model)

    assert abs(out["gx"] - out["gy"]) == pytest.approx(0)
    assert out["gx"] == pytest.approx(gcav_analytic(L, Rc1, Rc2))


@pytest.mark.parametrize(
    "L, Rc1, Rc2",
    (
        (1, 1, 0.9),
        (1, 1.9, 2.1),
        (1, np.inf, 2),
        (3994, 1934, 2245),
        (10e3, 5500, 5700),
        (10, 12, np.inf),
    ),
)
def test_asymm_fp_cav_round_trip_abcd(model: Model, fp_cavity_script, L, Rc1, Rc2):
    """Test detecting of round-trip ABCD for an asymmetric Fabry-Perot cavity, using
    analytics to compare results."""
    model.parse(fp_cavity_script)
    # Add a mode matched gauss at laser so that unstable cavity
    # values still run the file
    model.add_matched_gauss(model.L0.p1.o)
    model.parse("cp abcdx FP abcd x")
    model.parse("cp abcdy FP abcd y")
    model.ITM.Rc = -1 * Rc1
    model.ETM.Rc = Rc2
    model.elements["sc"].L = L

    A_analytic = lambda L, Rc2: 1 - 2 * L / Rc2
    B_analytic = lambda L, Rc2: 2 * L * (1 - L / Rc2)
    C_analytic = lambda L, Rc1, Rc2: 4 * L / (Rc1 * Rc2) - 2 / Rc1 - 2 / Rc2
    D_analytic = (
        lambda L, Rc1, Rc2: 1 + 4 * L * L / (Rc1 * Rc2) - 2 * L / Rc2 - 4 * L / Rc1
    )

    expected = np.array(
        [
            [A_analytic(L, Rc2), B_analytic(L, Rc2)],
            [C_analytic(L, Rc1, Rc2), D_analytic(L, Rc1, Rc2)],
        ]
    )

    out = noxaxis(model)

    assert np.all(np.abs(out["abcdx"] - out["abcdy"]) == pytest.approx(0))
    assert_allclose(out["abcdx"], expected)


@pytest.mark.parametrize(
    "L, Rc", ((1, 1), (1, 2), (1, np.inf), (4e3, 2090), (10e3, 5580),),
)
def test_symm_fp_cav_rtgouy(model: Model, fp_cavity_script, L, Rc):
    """Test detecting of round-trip Gouy phase for a symmetric Fabry-Perot cavity, using
    analytics to compare results."""
    model.parse(fp_cavity_script)
    # Add a mode matched gauss at laser so that unstable cavity
    # values still run the file
    model.add_matched_gauss(model.L0.p1.o)
    model.parse("cp xgouy FP gouy x")
    model.parse("cp ygouy FP gouy y")
    model.ITM.Rc = -1 * Rc
    model.ETM.Rc = Rc
    model.elements["sc"].L = L

    # For symm cav, psi_rt = 2 acos(g)
    gouy_analytic = lambda L, Rc: np.degrees(2 * np.arccos(1 - L / Rc))

    out = noxaxis(model)

    assert abs(out["xgouy"] - out["ygouy"]) == pytest.approx(0)
    assert out["xgouy"] == pytest.approx(gouy_analytic(L, Rc))


@pytest.mark.parametrize(
    "L, Rc1, Rc2",
    (
        (1, 1, 0.9),
        (1, 1.9, 2.1),
        (1, np.inf, 2),
        (3994, 1934, 2245),
        (10e3, 5500, 5700),
        (10, 12, np.inf),
    ),
)
def test_asymm_fp_cav_rtgouy(model: Model, fp_cavity_script, L, Rc1, Rc2):
    """Test detecting of round-trip Gouy phase for an asymmetric Fabry-Perot cavity,
    using analytics to compare results."""
    model.parse(fp_cavity_script)
    # Add a mode matched gauss at laser so that unstable cavity
    # values still run the file
    model.add_matched_gauss(model.L0.p1.o)
    model.parse("cp xgouy FP gouy x")
    model.parse("cp ygouy FP gouy y")
    model.ITM.Rc = -1 * Rc1
    model.ETM.Rc = Rc2
    model.elements["sc"].L = L

    gs_analytic = lambda L, Rc: 1 - L / Rc
    g1 = gs_analytic(L, Rc1)
    g2 = gs_analytic(L, Rc2)
    # psi_rt = 2 acos(sgn(B) sqrt(g1 g2)); where B = 2L(1 - L/Rc2) --- i.e. sgn(B) = sgn(g2)
    gouy_analytic = np.degrees(2 * np.arccos(np.sign(g2) * np.sqrt(g1 * g2)))

    out = noxaxis(model)

    assert abs(out["xgouy"] - out["ygouy"]) == pytest.approx(0)
    assert out["xgouy"] == pytest.approx(gouy_analytic)


@pytest.mark.parametrize(
    "L, Rc", ((1, 1), (1, 2), (1, np.inf), (4e3, 2090), (10e3, 5580),),
)
def test_symm_fp_cav_modeseparation(model: Model, fp_cavity_script, L, Rc):
    """Test detecting of mode separation frequency for a symmetric Fabry-Perot cavity,
    using analytics to compare results."""
    model.parse(fp_cavity_script)
    # Add a mode matched gauss at laser so that unstable cavity
    # values still run the file
    model.add_matched_gauss(model.L0.p1.o)
    model.parse("cp dfx FP modesep x")
    model.parse("cp dfy FP modesep y")
    model.ITM.Rc = -1 * Rc
    model.ETM.Rc = Rc
    model.elements["sc"].L = L

    fsr_analytic = lambda L: constants.C_LIGHT / (2 * L)
    gouy_analytic = lambda L, Rc: 2 * np.arccos(1 - L / Rc)

    # Analytic equation for mode separation frequency
    # is df = FSR * psi / (2 * pi)
    def df_analytic(L, Rc):
        gouy = gouy_analytic(L, Rc)
        # If round-trip gouy in upper quadrant (B < 0),
        # need to subtract df from FSR to get correct value
        if gouy > np.pi:
            return fsr_analytic(L) * (1 - 0.5 * gouy / np.pi)

        return 0.5 * fsr_analytic(L) * gouy / np.pi

    out = noxaxis(model)

    assert abs(out["dfx"] - out["dfy"]) == pytest.approx(0)
    assert out["dfx"] == pytest.approx(df_analytic(L, Rc))


@pytest.mark.parametrize(
    "L, Rc", ((1, 2), (4e3, 2090), (10e3, 5580),),
)
def test_symm_fp_cav_beamsize(model: Model, fp_cavity_script, L, Rc):
    """Test detecting of beam-size at mirrors for a symmetric Fabry-Perot cavity, using
    analytics to compare results."""
    model.parse(fp_cavity_script)
    model.parse("cp wx FP w x")
    model.parse("cp wy FP w y")
    model.ITM.Rc = -1 * Rc
    model.ETM.Rc = Rc
    model.elements["sc"].L = L

    wl = model.lambda0
    # Equation for beam size at cavity mirrors for symmetric cavity
    # in terms of the length L and mirror radii of curvature Rc
    w_analytic = lambda L, Rc: np.sqrt((wl / np.pi) * np.sqrt(Rc * L / (2 - L / Rc)))

    out = noxaxis(model)

    assert abs(out["wx"] - out["wy"]) == pytest.approx(0)
    assert out["wx"] == pytest.approx(w_analytic(L, Rc))


@pytest.mark.parametrize(
    "L, Rc1, Rc2",
    (
        (1, 1.9, 2.1),
        (1, np.inf, 2),
        (3994, 1934, 2245),
        (10e3, 5500, 5700),
        (10, 12, np.inf),
    ),
)
def test_asymm_fp_cav_beamsizes(model: Model, fp_cavity_script, L, Rc1, Rc2):
    """Test detecting of beam-sizes at mirrors for an asymmetric Fabry-Perot cavity,
    using analytics to compare results."""
    model.parse(fp_cavity_script)
    model.parse("cp wx1 FP w x")
    model.parse("cp wy1 FP w y")
    model.parse("bp wx2 ETM.p1.i w x")
    model.parse("bp wy2 ETM.p1.i w y")
    model.ITM.Rc = -1 * Rc1
    model.ETM.Rc = Rc2
    model.elements["sc"].L = L

    wl = model.lambda0
    # Equation for beam sizes at cavity mirrors for asymmetric cavity
    # in terms of the length L and mirror radii of curvature Rc1, Rc2
    gs_analytic = lambda L, Rc: 1 - L / Rc
    g1 = gs_analytic(L, Rc1)
    g2 = gs_analytic(L, Rc2)
    # w1 = sqrt((L\lambda / pi) sqrt(g2 / g1(1 - g1 g2)))
    w1_analytic = lambda L: np.sqrt(
        (L * wl / np.pi) * np.sqrt(g2 / (g1 * (1 - g1 * g2)))
    )
    # Beam size at end mirror is sqrt(g1 / g2) * w1
    w2_analytic = lambda L: w1_analytic(L) * np.sqrt(g1 / g2)

    out = noxaxis(model)

    assert abs(out["wx1"] - out["wy1"]) == pytest.approx(0)
    assert abs(out["wx2"] - out["wy2"]) == pytest.approx(0)

    assert out["wx1"] == pytest.approx(w1_analytic(L))
    assert out["wx2"] == pytest.approx(w2_analytic(L))


@pytest.mark.parametrize(
    "L, Rc", ((1, 2), (4e3, 2090), (10e3, 5580),),
)
def test_symm_fp_cav_waistsize(model: Model, fp_cavity_script, L, Rc):
    """Test detecting of waist-size for a symmetric Fabry-Perot cavity, using analytics
    to compare results."""
    model.parse(fp_cavity_script)
    model.parse("cp w0x FP w0 x")
    model.parse("cp w0y FP w0 y")
    model.ITM.Rc = -1 * Rc
    model.ETM.Rc = Rc
    model.elements["sc"].L = L

    wl = model.lambda0
    # Equation for waist size of symmetric cavity in terms
    # of the length L and mirror radii of curvature Rc
    w0_analytic = lambda L, Rc: np.sqrt(
        (L * wl / (2 * np.pi)) * np.sqrt(2 * Rc / L - 1)
    )

    out = noxaxis(model)

    assert abs(out["w0x"] - out["w0y"]) == pytest.approx(0)
    assert out["w0x"] == pytest.approx(w0_analytic(L, Rc))


@pytest.mark.parametrize(
    "L, Rc1, Rc2",
    (
        (1, 1.9, 2.1),
        (1, np.inf, 2),
        (3994, 1934, 2245),
        (10e3, 5500, 5700),
        (10, 12, np.inf),
    ),
)
def test_asymm_fp_cav_waistsize(model: Model, fp_cavity_script, L, Rc1, Rc2):
    """Test detecting of waist-size for an asymmetric Fabry-Perot cavity, using
    analytics to compare results."""
    model.parse(fp_cavity_script)
    model.parse("cp w0x FP w0 x")
    model.parse("cp w0y FP w0 y")
    model.ITM.Rc = -1 * Rc1
    model.ETM.Rc = Rc2
    model.elements["sc"].L = L

    wl = model.lambda0
    # Equation for waist size of asymmetric cavity in terms of
    # the length L and mirror radii of curvature Rc1, Rc2
    gs_analytic = lambda L, Rc: 1 - L / Rc
    g1 = gs_analytic(L, Rc1)
    g2 = gs_analytic(L, Rc2)
    # w0 = sqrt((L\lambda / pi) sqrt(g1 g2 (1 - g1 g2) / (g1 + g2 - 2g1 g2)^2))
    w0_analytic = lambda L: np.sqrt(
        (L * wl / np.pi)
        * np.sqrt(g1 * g2 * (1 - g1 * g2) / (g1 + g2 - 2 * g1 * g2) ** 2)
    )

    out = noxaxis(model)

    assert abs(out["w0x"] - out["w0y"]) == pytest.approx(0)
    assert out["w0x"] == pytest.approx(w0_analytic(L))


@pytest.mark.parametrize(
    "L, Rc", ((1, 2), (4e3, 2090), (10e3, 5580),),
)
def test_symm_fp_cav_waistpos(model: Model, fp_cavity_script, L, Rc):
    """Test detecting of waist position for a symmetric Fabry-Perot cavity, using
    analytics to compare results."""
    model.parse(fp_cavity_script)
    model.parse("cp zx FP z x")
    model.parse("cp zy FP z y")
    model.ITM.Rc = -1 * Rc
    model.ETM.Rc = Rc
    model.elements["sc"].L = L

    out = noxaxis(model)

    assert abs(out["zx"] - out["zy"]) == pytest.approx(0)
    # Factor of -1 because Finesse uses +z towards beam waist so
    # waist position should be at - L / 2 as measured from cavity
    # source node
    assert out["zx"] == pytest.approx(-1 * 0.5 * L)


@pytest.mark.parametrize(
    "L, Rc1, Rc2",
    (
        (1, 1.9, 2.1),
        (1, np.inf, 2),
        (3994, 1934, 2245),
        (10e3, 5500, 5700),
        (10, 12, np.inf),
    ),
)
def test_asymm_fp_cav_waistpos(model: Model, fp_cavity_script, L, Rc1, Rc2):
    """Test detecting of waist position for an asymmetric Fabry-Perot cavity, using
    analytics to compare results."""
    model.parse(fp_cavity_script)
    model.parse("cp zx FP z x")
    model.parse("cp zy FP z y")
    model.ITM.Rc = -1 * Rc1
    model.ETM.Rc = Rc2
    model.elements["sc"].L = L

    out = noxaxis(model)

    gs_analytic = lambda L, Rc: 1 - L / Rc
    g1 = gs_analytic(L, Rc1)
    g2 = gs_analytic(L, Rc2)
    # Equation for waist position, relative to ITM, for asymmetric cavity
    z_analytic = lambda L: -L * g2 * (1 - g1) / (g1 * (1 - g2) + g2 * (1 - g1))

    assert abs(out["zx"] - out["zy"]) == pytest.approx(0)
    # Factor of -1 because Finesse uses +z towards beam waist so
    # waist position should be at - L / 2 as measured from cavity
    # source node
    assert out["zx"] == pytest.approx(z_analytic(L))



"""
Outputs from a simulation / analysis run.

Listed below are all the sub-modules of the ``solutions`` module with
a brief description of the contents of each.
"""
from finesse.solutions.base import BaseSolution
from finesse.solutions.array import ArraySolution
from finesse.solutions.beamtrace import (
    ABCDSolution,
    PropagationSolution,
    AstigmaticPropagationSolution,
    BeamTraceSolution,
)

__all__ = (
    "BaseSolution",
    "ArraySolution",
    "ABCDSolution",
    "PropagationSolution",
    "AstigmaticPropagationSolution",
    "BeamTraceSolution",
)



import pytest
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler):
    cmd = lambda a=None, b=None, c=None, d=None: None
    compiler.spec.register_command(
        "fake_command", {"setter": cmd, "getter": cmd, "singular": True},
    )
    return compiler


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            "fake_command()",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """
            ),
            id="function with no arguments",
        ),
        pytest.param(
            "fake_command(1)",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'1'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """
            ),
            id="function with 1 argument",
        ),
        pytest.param(
            "fake_command(1, 2)",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ',', ' ', ')']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'1'"
                        type "VALUE"
                    ]
                    node [
                        id 3
                        label ".script.0.1"
                        token "'2'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 1
                        type "ARGUMENT"
                        order "1"
                    ]
                ]
                """
            ),
            id="function with 2 arguments",
        ),
        pytest.param(
            "fake_command(fake_command(3))",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 3
                        label ".script.0.0.0"
                        token "'3'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 2
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """,
            ),
            id="function with nested function argument",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_command(
                    fake_command(3)
                )
                """
            ),
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', '\\n', '    ', '\\n', ')']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 3
                        label ".script.0.0.0"
                        token "'3'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 2
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """,
            ),
            id="function with nested function argument across multiple lines",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_command(1)
                fake_command(2)
                """
            ),
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label ".script"
                        type "ROOT"
                        extra_tokens "['\\n']"
                    ]
                    node [
                        id 1
                        label ".script.0"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 2
                        label ".script.0.0"
                        token "'1'"
                        type "VALUE"
                    ]
                    node [
                        id 3
                        label ".script.1"
                        token "'fake_command'"
                        type "FUNCTION"
                        extra_tokens "['(', ')']"
                    ]
                    node [
                        id 4
                        label ".script.1.0"
                        token "'2'"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 0
                        type "ARGUMENT"
                        order "1"
                    ]
                    edge [
                        source 4
                        target 3
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """
            ),
            id="multiple functions",
        ),
    ),
)
def test_function(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)



"""
Motion amplitude and phase detector.
"""

import logging

import numpy as np

from finesse.detectors.general import Detector
from finesse.detectors.workspace import DetectorWorkspace

LOGGER = logging.getLogger(__name__)


class MotionDetectorWorkspace(DetectorWorkspace):
    def __init__(self, owner, sim):
        super().__init__(owner, sim)


class MotionDetector(Detector):
    """Represents a motion detector which calculates the amplitude and phase of surface motion.

    Parameters
    ----------
    name : str
        Name of newly created motion detector.

    node : :class:`.Node`
        Node to read output from.
    """

    def __init__(self, name, node):
        Detector.__init__(self, name, node, dtype=np.complex128, label="Motion")

    def _get_workspace(self, sim):
        def output_fn(ws):
            if ws.sim.signal:
                return ws.sim.signal.get_out(self.node, 0, 0) * ws.sim.model._x_scale

        ws = MotionDetectorWorkspace(self, sim)
        ws.set_output_fn(output_fn)
        return ws



"""Mode mismatch detector tests."""

import pytest
from finesse import Model
from finesse.analysis import xaxis
from finesse.components import Gauss
from finesse.detectors import BeamPropertyDetector
from finesse.gaussian import transform_beam_param, BeamParam
from numpy.testing import assert_allclose


@pytest.fixture
def matched_input_to_cavity_model():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    s s0 L0.p1 ITM.p1

    m ITM Rc=-2.5
    s sc ITM.p2 ETM.p1 L=1
    m ETM Rc=2.5

    cav FP ITM.p2

    mmd mmt1 ITM.p1 ITM.p2
    mmd mmt2 ITM.p2 ITM.p1
    cp fp_q FP q
    """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    # Add an input beam with q matched to cavity
    trace = IFO.beam_trace()
    IFO.add(Gauss("gL0", IFO.L0.p1.o, q=trace[IFO.L0.p1.o]))

    return IFO


def test_eigenmode_scan_with_etm_roc_changing(matched_input_to_cavity_model):
    """Test sweeping of ETM RoC gives mode mismatch with input beam."""
    IFO = matched_input_to_cavity_model
    out = xaxis(IFO.ETM.Rcx, "lin", 2.5, 10, 100)
    mmt1 = out["mmt1"]
    mmt2 = out["mmt2"]

    # First check that mismatch is equal for both
    # transmission couplings at ITM
    assert_allclose(mmt1, mmt2)

    q2s = out["fp_q"]
    q1 = IFO.beam_trace()[IFO.ITM.p1.i].qx
    q1p = transform_beam_param(IFO.ITM.abcd12x, q1)
    mms = 1.0 - BeamParam.overlap(q1p, q2s)

    assert_allclose(mmt1, mms)


def test_eigenmode_scan_with_itm_roc_changing(matched_input_to_cavity_model):
    """Test sweeping of ITM RoC gives mode mismatch with input beam."""
    IFO = matched_input_to_cavity_model
    out = xaxis(IFO.ITM.Rcx, "lin", 2.5, 10, 100)
    mmt1 = out["mmt1"]
    mmt2 = out["mmt2"]

    # First check that mismatch is equal for both
    # transmission couplings at ITM
    assert_allclose(mmt1, mmt2)

    q2s = out["fp_q"]
    q1 = IFO.beam_trace()[IFO.ITM.p1.i].qx
    q1p = transform_beam_param(IFO.ITM.abcd12x, q1)
    mms = 1.0 - BeamParam.overlap(q1p, q2s)

    assert_allclose(mmt1, mms)


def test_input_beam_waist_size_x_scan(matched_input_to_cavity_model):
    """Test sweeping of input beam waist size gives mode mismatch to cavity."""
    IFO = matched_input_to_cavity_model
    IFO.add(BeamPropertyDetector("q1", IFO.ITM.p1.i, "q"))
    out = xaxis(IFO.gL0.w0x, "lin", IFO.gL0.w0x.value, 2 * IFO.gL0.w0x.value, 100)
    mmt1 = out["mmt1"]
    mmt2 = out["mmt2"]

    # First check that mismatch is equal for both
    # transmission couplings at ITM
    assert_allclose(mmt1, mmt2)

    q2s = out["fp_q"]
    # Check that eigenmode hasn't changed
    assert_allclose(q2s, IFO.FP.qx.q)

    q1 = out["q1"]
    q1p = transform_beam_param(IFO.ITM.abcd12x, q1)
    mms = 1.0 - BeamParam.overlap(q1p, q2s)

    assert_allclose(mmt1, mms)


def test_input_beam_waist_size_y_scan(matched_input_to_cavity_model):
    """Test that sweeping input beam waist size in sagittal plane results in zeroes for
    mismatch detectors (as they should be detecting in tangential plane)"""
    IFO = matched_input_to_cavity_model
    out = xaxis(IFO.gL0.w0y, "lin", IFO.gL0.w0y.value, 2 * IFO.gL0.w0y.value, 100)
    mmt1 = out["mmt1"]
    mmt2 = out["mmt2"]

    # First check that mismatch is equal for both
    # transmission couplings at ITM
    assert_allclose(mmt1, mmt2)

    q2s = out["fp_q"]
    # Check that eigenmode hasn't changed
    assert_allclose(q2s, IFO.FP.qx.q)

    assert_allclose(mmt1, 0)


# TODO (sjr) Test mismatches at different components (e.g. beamsplitters)



"""Kat shell tools."""

import cmd
from collections import ChainMap

from . import parse
from .spec import KatSpec
from ..exceptions import FinesseException
from .. import PROGRAM, __version__, Model


class KatShell(cmd.Cmd):
    """Kat shell."""

    intro = f"Welcome to the {PROGRAM} {__version__} shell. Type help or ? to list commands.\n"
    prompt = "(kat) "
    doc_leader = "For further help on a topic, type 'help <topic>'.\n"
    doc_header = "Available elements:"
    misc_header = "Available commands ('!<command>'):"
    model = None
    solution = None

    def __init__(self, *args, **kwargs):
        spec = KatSpec()

        # Set kat script elements as shell commands.
        for element, adapter in spec.elements.items():
            setattr(self.__class__, f"do_{element}", self._wrap_kat_element(element))
            setattr(
                self.__class__,
                f"help_{element}",
                self._wrap_kat_directive_help(adapter),
            )

        # Set kat script command as magic methods.
        for command, adapter in ChainMap(spec.commands, spec.analyses).items():
            setattr(
                self.__class__, f"_magic_{command}", self._wrap_kat_command(command)
            )
            setattr(
                self.__class__,
                f"help_{command}",
                self._wrap_kat_directive_help(adapter),
            )

        # Set help for magic methods.
        for attr in dir(self):
            if attr.startswith("_magic_"):
                setattr(self.__class__, f"help_{attr[7:]}", self._wrap_magic_help(attr))

        super().__init__(*args, **kwargs)

    @classmethod
    def _wrap_kat_element(cls, element):
        def wrapper(*args):
            # Get rid of self, and add the element instead.
            shell = args[0]
            args = [element] + list(args[1:])

            original_line = " ".join(args)

            try:
                parse(original_line, model=shell.model)
            except FinesseException as e:
                print(e, file=shell.stdout)

        return wrapper

    @classmethod
    def _wrap_kat_command(cls, command):
        def wrapper(*args):
            # Get rid of self, and add the command instead.
            shell = args[0]
            args = [command] + list(args[1:])

            original_line = " ".join(args)

            try:
                parse(original_line, model=shell.model)
            except FinesseException as e:
                print(e, file=shell.stdout)

        return wrapper

    @classmethod
    def _wrap_kat_directive_help(cls, adapter):
        def wrapper(*args):
            shell = args[0]
            print(
                adapter.call_signature_type.__doc__, file=shell.stdout,
            )

        return wrapper

    @classmethod
    def _wrap_magic_help(cls, method):
        def wrapper(*args):
            shell = args[0]
            print(
                getattr(cls, method).__doc__, file=shell.stdout,
            )

        return wrapper

    def _reset(self):
        self.model = Model()

    def preloop(self):
        self._reset()

    def close(self):
        self.model = None
        self.solution = None

    def do_shell(self, arg):
        """Handler for commands starting with "!"."""
        magic_attr = f"_magic_{arg}"
        if hasattr(self, magic_attr):
            try:
                getattr(self, magic_attr)(arg)
            except FinesseException as e:
                print(e, file=self.stdout)
        else:
            try:
                parse(arg, model=self.model)
            except FinesseException as e:
                print(e, file=self.stdout)

    def _magic_model_info(self, arg):
        """Show information about the current model (if no argument is specified) or a particular \
        model element (if its name is specified as an argument)."""
        item = self.model.reduce_get_attr(arg) if arg else self.model
        print(f"{item.info()}", file=self.stdout)

    def _magic_model_get(self, arg):
        """Get a model element attribute."""
        print(self.model.reduce_get_attr(arg), file=self.stdout)

    def _magic_model_set(self, arg):
        """Set an attribute of the model or a model element."""
        self.model.reduce_set_attr(*arg.split())

    def _magic_model_run(self, arg):
        """Run the analysis."""
        try:
            self.solution = self.model.run()
        except FinesseException as e:
            print(e, file=self.stdout)

    def _magic_model_plot(self, arg):
        """Plot the solution."""
        self.solution.plot()

    def _magic_kat_load(self, arg):
        """Parse the specified kat file into the current model."""
        self.model.parse_file(arg)

    def _magic_kat_save(self, arg):
        """Save the current model to file."""
        self.model.unparse_file(arg)

    def _magic_model_reset(self, arg):
        """Reset the model."""
        self._reset()



import click
from .model import run
from .util import (
    set_debug,
    input_file_argument,
    plot_option,
    trace_option,
    legacy_option,
    verbose_option,
    quiet_option,
    fancy_error_option,
    debug_option,
    log_exclude_option,
)


@click.group()
@click.pass_context
def debug(ctx):
    """Debug tools."""
    # The debug flag is implied.
    set_debug(ctx, None, True)


@debug.command(name="run")
@input_file_argument
@plot_option
@trace_option
@legacy_option
@verbose_option
@quiet_option
@fancy_error_option
@debug_option
@log_exclude_option
@click.pass_context
def run_debug(
    ctx, input_file, plot, trace, legacy,
):
    """Run a Finesse script with Python fault handler enabled.

    If the file extension is 'py', it is interpreted as Python code; otherwise it is
    parsed assuming it to be kat script.
    """
    import os
    import faulthandler

    click.secho("Enabling Python fault handler", fg="yellow")
    faulthandler.enable()

    root, ext = os.path.splitext(input_file.name)
    if ext.casefold() == ".py":
        exec(input_file.read())
    else:
        # Forward arguments to normal "run" command.
        ctx.forward(run)


if __name__ == "__main__":
    run_debug()



"""CLI syntax command tests."""

import pytest
from finesse.__main__ import syntax
from finesse.script.spec import KatSpec
from .util import sanitized_output


@pytest.mark.parametrize("directive", KatSpec().directives)
def test_syntax(cli, directive):
    """Test syntax output."""
    # Search for the specified directive.
    cli_result = cli.invoke(syntax, [directive])
    output = sanitized_output(cli_result)
    assert cli_result.exit_code == 0
    assert directive in output



import pytest
from finesse.script.graph import KatNodeType, KatEdgeType
from ..util import (
    NAME,
    NUMBER,
    EQUALS,
    COMMA,
    SPACE,
    LPAREN,
    RPAREN,
    LBRACKET,
    RBRACKET,
)


token_value_parameters = pytest.mark.parametrize(
    "token",
    (
        # Numbers.
        NUMBER(1, 1, "1"),
        # Parameters.
        NAME(1, 1, "m1.T"),
        NAME(1, 1, "&m1.T"),
        # Constants.
        NAME(1, 1, "A"),
        # Keywords.
        NAME(1, 1, "lin"),
    ),
)


@token_value_parameters
def test_value(unfiller, graph, token):
    """Single values."""
    graph.add_node("myval", token=token, type=KatNodeType.VALUE)
    assert unfiller.unfill("myval", graph) == token.raw_value


@pytest.mark.parametrize(
    "key_token,equals_token,value_token",
    (
        # Numbers.
        (NAME(1, 1, "a"), EQUALS(1, 2), NUMBER(1, 3, "1")),
        # Parameters.
        (NAME(1, 1, "a"), EQUALS(1, 2), NAME(1, 3, "m1.T")),
        (NAME(1, 1, "a"), EQUALS(1, 2), NAME(1, 3, "&m1.T")),
        # Constants.
        (NAME(1, 1, "a"), EQUALS(1, 2), NAME(1, 3, "A")),
        # Keywords.
        (NAME(1, 1, "a"), EQUALS(1, 2), NAME(1, 3, "lin")),
    ),
)
def test_kwarg(unfiller, graph, key_token, equals_token, value_token):
    """Keyword values in arguments."""
    graph.add_node(
        "myarg",
        type=KatNodeType.VALUE,
        token=value_token,
        key_token=key_token,
        extra_tokens=[equals_token],
    )
    expected = "".join(
        [tok.raw_value for tok in (key_token, equals_token, value_token)]
    )
    assert unfiller.unfill("myarg", graph) == expected


@pytest.mark.parametrize(
    "operator,arguments,extra,expected",
    (
        pytest.param(
            NAME(1, 2, "+"),
            [NUMBER(1, 1, "1"), NUMBER(1, 3, "2")],
            [],
            "1+2",
            id="1+2",
        ),
    ),
)
def test_expression(unfiller, graph, operator, arguments, extra, expected):
    expression_node = "myexpr"
    graph.add_node(
        expression_node, token=operator, type=KatNodeType.EXPRESSION, extra_tokens=extra
    )
    for order, argument in enumerate(arguments):
        arg_node = f"{expression_node}.{order}"
        graph.add_node(arg_node, token=argument, type=KatNodeType.VALUE)
        graph.add_edge(
            arg_node, expression_node, type=KatEdgeType.ARGUMENT, order=order
        )

    assert unfiller.unfill(expression_node, graph) == expected


@pytest.mark.parametrize(
    "function,arguments,extra,expected",
    (
        pytest.param(
            NAME(1, 1, "cos"),
            [NUMBER(1, 5, "1")],
            [LPAREN(1, 4), RPAREN(1, 6)],
            "cos(1)",
            id="cos(1)",
        ),
        pytest.param(
            NAME(1, 1, "cos"),
            [NUMBER(1, 5, "1"), NUMBER(1, 8, "2")],
            [LPAREN(1, 4), COMMA(1, 6), RPAREN(1, 9)],
            "cos(1,2)",
            id="cos(1,2)",
        ),
        pytest.param(
            NAME(1, 1, "cos"),
            [NUMBER(1, 5, "1"), NUMBER(1, 8, "2")],
            [LPAREN(1, 4), COMMA(1, 6), SPACE(1, 7), RPAREN(1, 9)],
            "cos(1, 2)",
            id="cos(1, 2)",
        ),
    ),
)
def test_expression_function(unfiller, graph, function, arguments, extra, expected):
    function_node = "myfunc"
    graph.add_node(
        function_node, token=function, type=KatNodeType.FUNCTION, extra_tokens=extra
    )
    for order, argument in enumerate(arguments):
        arg_node = f"{function_node}.{order}"
        graph.add_node(arg_node, token=argument, type=KatNodeType.VALUE)
        graph.add_edge(arg_node, function_node, type=KatEdgeType.ARGUMENT, order=order)

    assert unfiller.unfill(function_node, graph) == expected


@pytest.mark.parametrize(
    "arguments,extra,expected",
    (
        pytest.param(
            [NUMBER(1, 2, "1")], [LPAREN(1, 1), RPAREN(1, 3)], "(1)", id="(1)",
        ),
    ),
)
def test_grouped_expression(unfiller, graph, arguments, extra, expected):
    group_node = "mygroup"
    graph.add_node(group_node, type=KatNodeType.GROUPED_EXPRESSION, extra_tokens=extra)
    for order, argument in enumerate(arguments):
        arg_node = f"{group_node}.{order}"
        graph.add_node(arg_node, token=argument, type=KatNodeType.VALUE)
        graph.add_edge(arg_node, group_node, type=KatEdgeType.ARGUMENT, order=order)

    assert unfiller.unfill(group_node, graph) == expected


@pytest.mark.parametrize(
    "arguments,extra,expected",
    (
        pytest.param(
            [NUMBER(1, 2, "1")], [LBRACKET(1, 1), RBRACKET(1, 3)], "[1]", id="[1]",
        ),
        pytest.param(
            [NUMBER(1, 2, "1"), NUMBER(1, 5, "2")],
            [LBRACKET(1, 1), COMMA(1, 3), SPACE(1, 4), RBRACKET(1, 6)],
            "[1, 2]",
            id="[1, 2]",
        ),
    ),
)
def test_array(unfiller, graph, arguments, extra, expected):
    array_node = "myarray"
    graph.add_node(array_node, type=KatNodeType.ARRAY, extra_tokens=extra)
    for order, argument in enumerate(arguments):
        arg_node = f"{array_node}.{order}"
        graph.add_node(arg_node, token=argument, type=KatNodeType.VALUE)
        graph.add_edge(arg_node, array_node, type=KatEdgeType.ARGUMENT, order=order)

    assert unfiller.unfill(array_node, graph) == expected


@pytest.mark.parametrize(
    "directive,element_name,arguments,extra,expected",
    (
        pytest.param(
            NAME(1, 1, "fake_element"),
            NAME(1, 14, "myelement"),
            [],
            [SPACE(1, 13)],
            "fake_element myelement",
            id="fake_element myelement",
        ),
        pytest.param(
            NAME(1, 1, "fake_element"),
            NAME(1, 14, "myelement"),
            [NUMBER(1, 24, "1")],
            [SPACE(1, 13), SPACE(1, 23)],
            "fake_element myelement 1",
            id="fake_element myelement 1",
        ),
        pytest.param(
            NAME(1, 1, "fake_element"),
            NAME(1, 14, "myelement"),
            [NUMBER(1, 24, "1"), NUMBER(1, 26, "2")],
            [SPACE(1, 13), SPACE(1, 23), SPACE(1, 25)],
            "fake_element myelement 1 2",
            id="fake_element myelement 1 2",
        ),
    ),
)
def test_element(unfiller, graph, directive, element_name, arguments, extra, expected):
    element_node = "myelement"
    graph.add_node(
        element_node,
        token=directive,
        name_token=element_name,
        type=KatNodeType.ELEMENT,
        extra_tokens=extra,
    )
    for order, argument in enumerate(arguments):
        arg_node = f"{element_node}.{order}"
        graph.add_node(arg_node, token=argument, type=KatNodeType.VALUE)
        graph.add_edge(arg_node, element_node, type=KatEdgeType.ARGUMENT, order=order)

    assert unfiller.unfill(element_node, graph) == expected


@pytest.mark.parametrize(
    "directive,arguments,extra,expected",
    (
        pytest.param(
            NAME(1, 1, "fake_function"),
            [],
            [LPAREN(1, 14), RPAREN(1, 15)],
            "fake_function()",
            id="fake_function()",
        ),
        pytest.param(
            NAME(1, 1, "fake_function"),
            [NUMBER(1, 15, "1")],
            [LPAREN(1, 14), RPAREN(1, 16)],
            "fake_function(1)",
            id="fake_function(1)",
        ),
        pytest.param(
            NAME(1, 1, "fake_function"),
            [NUMBER(1, 15, "1"), NUMBER(1, 18, "2")],
            [LPAREN(1, 14), COMMA(1, 16), SPACE(1, 17), RPAREN(1, 19)],
            "fake_function(1, 2)",
            id="fake_function(1, 2)",
        ),
    ),
)
def test_function(unfiller, graph, directive, arguments, extra, expected):
    function_node = "myfunction"
    graph.add_node(
        function_node, token=directive, type=KatNodeType.FUNCTION, extra_tokens=extra
    )
    for order, argument in enumerate(arguments):
        arg_node = f"{function_node}.{order}"
        graph.add_node(arg_node, token=argument, type=KatNodeType.VALUE)
        graph.add_edge(arg_node, function_node, type=KatEdgeType.ARGUMENT, order=order)

    assert unfiller.unfill(function_node, graph) == expected



"""Tests for the tokenizer without any normalization."""

import pytest
from ...util import dedent_multiline
from ..util import (
    NAME,
    NUMBER,
    STRING,
    BOOLEAN,
    PLUS,
    EQUALS,
    LPAREN,
    RPAREN,
    ENDMARKER,
    NEWLINE,
    SPACE,
    COMMENT,
    IMPLICITLINEEND,
)


def test_implicit_newline(tokenizer):
    """If there is no newline at the end of the script, an implicit newline is added."""
    assert list(tokenizer.tokenize("a")) == [
        NAME(1, 1, "a"),
        IMPLICITLINEEND(1, 2),
        ENDMARKER(2),
    ]


def test_explicit_newline(tokenizer):
    """If there is a newline at the end of the script, an implicit newline is NOT added."""
    assert list(tokenizer.tokenize("a\n")) == [
        NAME(1, 1, "a"),
        NEWLINE(1, 2),
        ENDMARKER(2),
    ]


@pytest.mark.parametrize(
    "string,expected",
    (
        ## Single line.
        (
            "1 + 1",
            [
                NUMBER(1, 1, "1"),
                SPACE(1, 2),
                PLUS(1, 3),
                SPACE(1, 4),
                NUMBER(1, 5, "1"),
                IMPLICITLINEEND(1, 6),
                ENDMARKER(2),
            ],
        ),
        ("true", [BOOLEAN(1, 1, "true"), IMPLICITLINEEND(1, 5), ENDMARKER(2),]),
        ("false", [BOOLEAN(1, 1, "false"), IMPLICITLINEEND(1, 6), ENDMARKER(2),]),
        (
            "a = ''",
            [
                NAME(1, 1, "a"),
                SPACE(1, 2),
                EQUALS(1, 3),
                SPACE(1, 4),
                STRING(1, 5, "''"),
                IMPLICITLINEEND(1, 7),
                ENDMARKER(2),
            ],
        ),
        (
            "a = 'foo'",
            [
                NAME(1, 1, "a"),
                SPACE(1, 2),
                EQUALS(1, 3),
                SPACE(1, 4),
                STRING(1, 5, "'foo'"),
                IMPLICITLINEEND(1, 10),
                ENDMARKER(2),
            ],
        ),
        (
            'a = "foo"',
            [
                NAME(1, 1, "a"),
                SPACE(1, 2),
                EQUALS(1, 3),
                SPACE(1, 4),
                STRING(1, 5, '"foo"'),
                IMPLICITLINEEND(1, 10),
                ENDMARKER(2),
            ],
        ),
        ("'\"'", [STRING(1, 1, "'\"'"), IMPLICITLINEEND(1, 4), ENDMARKER(2),]),
        ('"\'"', [STRING(1, 1, '"\'"'), IMPLICITLINEEND(1, 4), ENDMARKER(2),]),
        (
            # This is a weird one: "shrink" should be interpreted as an ID.
            '"doesn\'t "shrink", does it"',
            [
                STRING(1, 1, '"doesn\'t "'),
                NAME(1, 11, "shrink"),
                STRING(1, 17, '", does it"'),
                IMPLICITLINEEND(1, 28),
                ENDMARKER(2),
            ],
        ),
        ## Multi line.
        (
            dedent_multiline(
                """
                laser l1 (
                    # some comment
                    P=5
                )
                """
            ),
            [
                NAME(1, 1, "laser"),
                SPACE(1, 6),
                NAME(1, 7, "l1"),
                SPACE(1, 9),
                LPAREN(1, 10),
                NEWLINE(1, 11),
                SPACE(2, 1, 4),
                COMMENT(2, 5, "# some comment"),
                NEWLINE(2, 19),
                SPACE(3, 1, 4),
                NAME(3, 5, "P"),
                EQUALS(3, 6),
                NUMBER(3, 7, "5"),
                NEWLINE(3, 8),
                RPAREN(4, 1),
                IMPLICITLINEEND(4, 2),
                ENDMARKER(5),
            ],
        ),
    ),
)
def test_lines(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected



import logging

from finesse.parameter import Parameter
import finesse.analysis.actions as ac

LOGGER = logging.getLogger(__name__)


def noxaxis(model):
    analysis = ac.Noxaxis()
    return analysis.run(model)


def xaxis(
    param: Parameter,
    mode: str,
    start: float,
    stop: float,
    steps: int,
    relative: bool = False,
):
    """Runs a model to scan a parameter between two points for a number of steps.

    The model that is run is retrieved from the parameter reference.

    This should provide an equivalent to the xaxis command in Finesse v2.

    Parameters
    ----------
    param : :class:`.Parameter`
        Parameter of component to scan.
    mode : str
        'lin' or 'log' for linear or logarithmic step sizes.
    start, stop : float
        Start and end values of the scan.
    steps : int
        Number of steps between start and end.
    relative : float, optional
        Changes relative to parameter initial value
    """
    analysis = ac.Xaxis(
        param.full_name, mode, start, stop, steps, relative=relative
    )
    sol = analysis.run(param.owner._model)
    return sol


def x2axis(
    param1,
    mode1,
    start1,
    stop1,
    steps1,
    param2,
    mode2,
    start2,
    stop2,
    steps2,
    relative=0,
    **kwargs,
):
    """Runs a model to scan a parameter between two points for a number of steps.
    The model that is run is retrieved from the parameter reference.

    This should provide an equivalent to the x2axis command in Finesse v2.

    Parameters
    ----------
    param1, param2 : :class:`.Parameter`
        Parameter of component to scan.
    mode1, mode2 : str
        'lin' or 'log' for linear or logarithmic step sizes for axis 1 and 2.
    start1, stop1, start2, stop2 : float
        Start and end values of the scan.
    steps1, steps2 : int
        Number of steps between start and end.
    relative : float, optional
        Changes relative to parameter initial value

    Notes
    -----
    `param2` is in the inner loop and `param1` in the outer loop.
    """
    if param1.owner._model is not param2.owner._model:
        raise ValueError(
            f"Parameters {param1} and {param2}, are from different models."
        )

    analysis = ac.X2axis(
        f"{param1.owner.name}.{param1.name}",
        mode1,
        start1,
        stop1,
        steps1,
        f"{param2.owner.name}.{param2.name}",
        mode2,
        start2,
        stop2,
        steps2,
        relative=relative
    )

    return analysis.run(param1.owner._model)


def x3axis(
    param1,
    mode1,
    start1,
    stop1,
    steps1,
    param2,
    mode2,
    start2,
    stop2,
    steps2,
    param3,
    mode3,
    start3,
    stop3,
    steps3,
    relative=False,
    **kwargs,
):
    """Runs a model to scan a parameter between two points for a number of steps.
    The model that is run is retrieved from the parameter reference.

    This should provide an equivalent to the x3axis command in Finesse v2.

    Parameters
    ----------
    param1, param2, param3 : :class:`.Parameter`
        Parameter of component to scan.
    mode1, mode2, mode3 : str
        'lin' or 'log' for linear or logarithmic step sizes for each axis.
    start1, stop1, start2, stop2, start3, stop3 : float
        Start and end values of the scan.
    steps1, steps2, steps3 : int
        Number of steps between start and end.
    relative : float, optional
        Changes relative to parameter initial value

    Notes
    -----
    `param3` is in the inner loop and `param1` in the outer loop.
    """
    model = param1.owner._model
    if any(param.owner._model is not model for param in (param1, param2, param3)):
        raise ValueError(
            f"Parameters {param1}, {param2}, and {param3}, are not from the same models."
        )
    analysis = ac.X3axis(
        param1.full_name,
        mode1,
        start1,
        stop1,
        steps1,
        param2.full_name,
        mode2,
        start2,
        stop2,
        steps2,
        param3.full_name,
        mode3,
        start3,
        stop3,
        steps3,
        relative=relative
    )

    return analysis.run(param1.owner._model)



"""Controlling an interferometer via error signals."""

import logging

import finesse.detectors as detectors
from finesse.element import ModelElement

LOGGER = logging.getLogger(__name__)


class Lock(ModelElement):
    """A simple lock which computes and applies the feedback to a given parameter using an error
    signal.

    Parameters
    ----------
    name : str
        Name of newly created lock.

    errsig : Any
        An error signal parameter or an object capable of producing a real-type error signal. This
        is typically a demodulated :class:`.PowerDetector` instance (or the name of the instance).

    feedback_to : :class:`.Parameter`
        A parameter of the model to apply the locks' feedback signal to.

    gain : float
        Control loop gain.

    accuracy : float
        Threshold to decide whether the loop is locked.

    disabled : boolean
        If true this lock will not run when the `RunLocks()` action is used. Explicitly specifying 
        the name of the lock will override this setting, e.g. `RunLocks(name)`.

    offset : float
        An offset that is applied to the error signal before it is used.
    """

    def __init__(self, name, errsig, feedback_to, gain, accuracy, *, disabled=False, offset=0):
        super().__init__(name)

        self.__errsig = errsig
        self.__feedback_to = feedback_to
        self.__gain = gain
        self.__accuracy = accuracy
        self.offset = offset
        self.disabled = disabled

    def _on_add(self, model):
        if isinstance(self.__errsig, str):
            self.__errsig = self._model.elements[self.__errsig]
        self.__from_detector = isinstance(self.__errsig, detectors.Detector)

    @property
    def error_signal(self):
        """The error signal of the lock.
        """
        return self.__errsig

    @property
    def feedback(self):
        """A handle to the parameter which the feedback signal is applied to."""
        return self.__feedback_to

    @property
    def gain(self):
        return self.__gain

    @gain.setter
    def gain(self, value):
        self.__gain = float(value)

    @property
    def accuracy(self):
        return self.__accuracy

    @accuracy.setter
    def accuracy(self, value):
        self.__accuracy = float(value)



"""Tests checking that symbolic equations are set-up correctly during simulations."""

import numpy as np
from numpy.testing import assert_allclose
from finesse import Model


def test_mirror_refl_trns_self_ref():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    link(L0, ITM)
    m ITM R=1 T=1-&ITM.R

    pd refl ITM.p1.o
    pd trns ITM.p2.o

    xaxis(ITM.R, lin, 0, 1, 2)
    """
    )

    out = IFO.run()

    assert_allclose(out["refl"], np.array([0, 0.5, 1]))
    assert_allclose(out["trns"], np.array([1, 0.5, 0]))


def test_mirror_refl_trns_via_variable():
    IFO = Model()
    IFO.parse(
        """
    l L0 P=1
    link(L0, ITM)
    m ITM R=&v T=1-&v

    pd refl ITM.p1.o
    pd trns ITM.p2.o

    var v 1
    xaxis(v.value, lin, 0, 1, 2)
    """
    )

    out = IFO.run()

    assert_allclose(out["refl"], np.array([0, 0.5, 1]))
    assert_allclose(out["trns"], np.array([1, 0.5, 0]))
